import autoray as ar
import pytest
from numpy.testing import assert_allclose

import symmray as sr


@pytest.mark.parametrize("dual", (False, True))
def test_block_index_basics(dual):
    ix = sr.BlockIndex(
        chargemap={
            -2: 3,
            0: 1,
            1: 2,
        },
        dual=dual,
    )
    ix.check()
    assert ix.size_total == 6
    assert ix.num_charges == 3
    assert ix.size_of(-2) == 3
    assert ix.matches(ix.conj())
    assert repr(ix)


def test_z2symmetric_array_basics():
    x = sr.utils.get_rand("Z2", (3, 4, 5, 6))
    x.check()
    assert x.shape == (3, 4, 5, 6)
    assert x.ndim == 4
    assert x.num_blocks == 8
    assert x.get_sparsity() == 1
    x.test_allclose(x.copy())


all_symmetries = ("Z2", "Z3", "Z5", "U1", "Z2Z2", "U1U1")
multi_symmetries = ("Z2Z2", "U1U1")


@pytest.mark.parametrize("symmetry", all_symmetries)
def test_AbelianArray_to_dense(symmetry):
    x = sr.utils.get_rand(symmetry, (3, 4, 5, 6))
    assert ar.do("linalg.norm", x) == pytest.approx(
        ar.do("linalg.norm", x.to_dense())
    )
    assert ar.do("linalg.norm", 2 * x) == pytest.approx(
        ar.do("linalg.norm", 2 * x.to_dense())
    )
    assert_allclose(
        x.transpose((3, 1, 2, 0)).to_dense(),
        x.to_dense().transpose((3, 1, 2, 0)),
    )


@pytest.mark.parametrize("symmetry", all_symmetries)
@pytest.mark.parametrize("missing", (False, True))
@pytest.mark.parametrize("mode", ["insert", "concat"])
def test_AbelianArray_fuse(symmetry, missing, mode):
    x = sr.utils.get_rand(symmetry, (3, 4, 5, 6))

    if missing:
        sector_to_drop = list(x.sectors)[len(x.sectors) // 2]
        x.del_block(sector_to_drop)

    xf = x.fuse((0, 2), (1, 3), mode=mode)
    if symmetry == "Z2":
        assert xf.shape == (15, 24)
        assert xf.num_blocks == 2
    xu = xf.unfuse_all().transpose((0, 2, 1, 3))
    x.test_allclose(xu)


@pytest.mark.parametrize("symmetry", all_symmetries)
@pytest.mark.parametrize(
    "shape0, shape1",
    [
        ((3, 4, 5, 6), (3, 4, 5, 6)),
        ((3, 4, 5, 6), (12, 30)),
        ((3, 4, 5, 6), (3, 20, 6)),
        ((2, 2, 2, 2, 2), (4, 2, 4)),
        ((2, 2, 2, 1, 2, 2), (1, 4, 2, 4, 1)),
        ((1, 1, 1, 1), (1, 1)),
        ((1, 1, 1), (1, 1, 1)),
        ((1, 1), (1, 1, 1, 1)),
    ],
)
@pytest.mark.parametrize("seed", range(5))
def test_AbelianArray_reshape(symmetry, shape0, shape1, seed):
    x = sr.utils.get_rand(symmetry, shape0, seed=seed, subsizes="maximal")
    y = ar.do("reshape", x, shape1)
    assert all(da <= db for da, db in zip(y.shape, shape1))
    y.check()
    z = ar.do("reshape", y, shape0)
    z.check()
    x.test_allclose(z)


@pytest.mark.parametrize("symmetry", all_symmetries)
@pytest.mark.parametrize("seed", range(5))
@pytest.mark.parametrize("mode", ["insert", "concat"])
def test_abelian_reshape_unfuse(symmetry, seed, mode):
    a = sr.utils.get_rand(
        symmetry, (3, 4, 5, 6), seed=seed, subsizes="maximal"
    )
    b = a.fuse((0, 1), (2, 3), mode=mode)
    c = b.reshape((12, 5, 6))
    c.test_allclose(a.fuse((0, 1), mode=mode))


@pytest.mark.parametrize("symmetry", all_symmetries)
def test_fuse_conj_unfuse(symmetry):
    if symmetry not in multi_symmetries:
        charge = 1
    else:
        charge = (1, 1)

    d = 7
    x = sr.utils.get_rand(
        symmetry,
        (d, d),
        charge=charge,
        subsizes="maximal",
    )
    xf = x.fuse((0, 1))
    xfc = xf.conj()
    xfc.check()
    xfcu = xfc.unfuse_all()
    xfcu.check()
    xfcuc = xfcu.conj()
    xfcuc.check()
    x.test_allclose(xfcuc)


def test_calc_reshape_args_edgecase():
    from symmray.abelian_common import calc_reshape_args

    axs_unfuse, axs_fuse_groupings, axs_expand = calc_reshape_args(
        shape=(4, 4, 4), newshape=(16, 4), subshapes=(None, None, (4, 4))
    )
    assert axs_unfuse == ()
    assert axs_expand == ()
    assert axs_fuse_groupings == (((0, 1),),)


@pytest.mark.parametrize("symmetry", all_symmetries)
@pytest.mark.parametrize("subsizes", ("maximal", "equal"))
@pytest.mark.parametrize("mode", ("fused", "blockwise"))
@pytest.mark.parametrize(
    "shape1,shape2,axes",
    [
        ((10,), (10,), 1),
        ((4, 5), (4, 5), 2),
        ((4,), (5,), 0),
    ],
)
def test_tensordot(symmetry, shape1, shape2, axes, subsizes, mode):
    if symmetry not in multi_symmetries:
        charge = 1
    else:
        charge = (1, 1)

    a = sr.utils.get_rand(
        symmetry,
        shape1,
        duals=[False] * len(shape1),
        charge=charge,
        subsizes=subsizes,
    )
    b = sr.utils.get_rand(
        symmetry,
        shape2,
        duals=[True] * len(shape2),
        charge=charge,
        subsizes=subsizes,
    )

    with sr.default_tensordot_mode(mode):
        c = ar.do("tensordot", a, b, axes=axes)
        assert sr.get_default_tensordot_mode() == mode
    assert sr.get_default_tensordot_mode() == "auto"

    d = ar.do("tensordot", a.to_dense(), b.to_dense(), axes=axes)

    if isinstance(c, sr.AbelianArray):
        if symmetry == "Z2":
            assert c.charge == 0
        elif symmetry == "U1":
            assert c.charge == 2
        elif symmetry == "Z2Z2":
            assert c.charge == (0, 0)
        elif symmetry == "U1U1":
            assert c.charge == (2, 2)
        if c.shape == d.shape:
            # XXX: can only check when no empty charges have been dropped
            assert_allclose(c.to_dense(), d)
        elif c.num_blocks > 0:
            # misaligned charges are all zero entries -> check sum
            assert_allclose(d.sum(), c.sum())
    else:
        assert_allclose(c, d)


@pytest.mark.parametrize("symm", all_symmetries)
def test_tensordot_fused_with_already_fused_arrays(symm):
    a, b, c, d = (sr.utils.rand_index(symm, d) for d in [4, 3, 2, 5])
    x = sr.utils.get_rand(symm, (a, b, c))
    y = sr.utils.get_rand(symm, (c.conj(), d))
    z = sr.tensordot(x, y, axes=[(2,), (0,)]).fuse((0, 1))
    xf = x.fuse((0, 1))
    zf = sr.tensordot(xf, y, axes=[(1,), (0,)])
    zf.test_allclose(z)


@pytest.mark.parametrize("symmetry", all_symmetries)
def test_AbelianArray_reductions(symmetry):
    x = sr.utils.get_rand(symmetry, (3, 4, 5, 6), dist="uniform", seed=42)
    assert ar.do("min", x) < ar.do("max", x) < ar.do("sum", x)


@pytest.mark.parametrize("symmetry", all_symmetries)
def test_block_multiply_diagonal(symmetry):
    import numpy as np

    rng = np.random.default_rng(42)
    x = sr.utils.get_rand(symmetry, (3, 4, 5, 6))
    axis = 2
    v = sr.BlockVector(
        {c: rng.normal(size=d) for c, d in x.indices[axis].chargemap.items()}
    )
    y = ar.do("multiply_diagonal", x, v, axis=axis)
    yd = y.to_dense()

    assert_allclose(
        yd,
        np.einsum("abcd,c->abcd", x.to_dense(), v.to_dense()),
    )


@pytest.mark.parametrize(
    "eq",
    (
        "ab->ab",
        "aa->",
        "aabb->",
        "abab->",
        "cba->abc",
        "cbabc->a",
    ),
)
@pytest.mark.parametrize("symmetry", all_symmetries)
def test_einsum_single_term(eq, symmetry):
    lhs, rhs = eq.split("->")

    indices = {
        "a": sr.utils.rand_index(symmetry, 3),
        "b": sr.utils.rand_index(symmetry, 4),
        "c": sr.utils.rand_index(symmetry, 5),
        "d": sr.utils.rand_index(symmetry, 6),
    }

    shape = []
    seen = set()
    for q in lhs:
        if q in seen:
            shape.append(indices[q].conj())
        else:
            shape.append(indices[q])
            seen.add(q)

    x = sr.utils.get_rand(symmetry, shape)
    x.check()
    dx = x.to_dense()
    y = ar.do("einsum", eq, x)
    if rhs:
        y.check()
        dy = y.to_dense()
    else:
        dy = y
    assert_allclose(dy, ar.do("einsum", eq, dx))


@pytest.mark.parametrize("symm", ["Z2", "U1", "Z3"])
def test_can_pickle_abelian_array(symm):
    import pickle
    import tempfile

    # XXX: add delete=False to avoid win github action permission error
    with tempfile.NamedTemporaryFile(suffix=".pkl", delete=False) as tmpf:
        tmp_fname = tmpf.name
        # create dynamic symmetry
        symm = sr.get_symmetry(symm)
        # save to disk
        with open(tmp_fname, "wb") as f:
            pickle.dump(symm, f)
        # load from disk
        with open(tmp_fname, "rb") as f:
            symm_loaded = pickle.load(f)
        assert symm == symm_loaded

    with tempfile.NamedTemporaryFile(suffix=".pkl", delete=False) as tmpf:
        tmp_fname = tmpf.name
        # create dynamic symmetry
        x = sr.utils.get_rand(symm, (2, 3, 4))
        # save to disk
        with open(tmp_fname, "wb") as f:
            pickle.dump(x, f)
        # load from disk
        with open(tmp_fname, "rb") as f:
            x_loaded = pickle.load(f)
        x.test_allclose(x_loaded)


@pytest.mark.parametrize("symmetry", all_symmetries)
@pytest.mark.parametrize("ndim", [1, 2, 3, 4])
@pytest.mark.parametrize("seed", range(10))
def test_abelian_array_slice(symmetry, ndim, seed):
    import numpy as np

    rng = sr.utils.get_rng(seed)
    shape = tuple(map(int, rng.integers(6, 10, size=ndim)))
    x = sr.utils.get_rand(symmetry, shape, seed=seed)

    for ax in range(ndim):
        for d in range(shape[ax]):
            selector = (
                (slice(None),) * ax + (d,) + (slice(None),) * (ndim - ax - 1)
            )
            np.testing.assert_allclose(
                x.to_dense()[selector],
                x[selector].to_dense(),
            )



"""Sparse backend block vector class."""

import autoray as ar

from ..common import SymmrayCommon
from ..vector_common import VectorCommon
from .sparse_data_common import BlockCommon


class BlockVector(BlockCommon, VectorCommon, SymmrayCommon):
    """Class for storing block vectors with sparse storage, e.g. for the
    singular- or eigen- values of a matrix.

    This is equivalent to the diagonal of a zero charge abelian matrix.

    Parameters
    ----------
    blocks : dict
        A dictionary mapping sector keys to array blocks.
    """

    __slots__ = ("_blocks",)
    ndim = 1

    def __init__(self, blocks):
        self._init_blockcommon(blocks)

    def copy(self):
        return self._copy_blockcommon()

    @property
    def size(self):
        """The total size of all elements in the vector."""
        # compute lazily
        _size = ar.get_lib_fn(self.backend, "size")
        return sum(_size(x) for x in self.get_all_blocks())

    @property
    def shape(self):
        """Get the effective shape of the vector."""
        return (self.size,)

    def _binary_blockwise_op(self, other, fn, missing=None, inplace=False):
        return self._binary_blockwise_op_abelian(
            other, fn, missing=missing, inplace=inplace
        )

    def check(self):
        """Check that the block vector is well formed."""
        ndims = {ar.ndim(x) for x in self.get_all_blocks()}
        if len(ndims) != 1:
            raise ValueError(f"blocks have different ndims: {ndims}")
        assert self.size == sum(ar.size(s) for s in self.get_all_blocks())

    def to_flat(self):
        """Convert the block sparse vector to a flat vector."""
        from ..flat.flat_vector import FlatVector

        return FlatVector.from_blocks(self.blocks)

    def to_dense(self):
        """Convert the block vector to a dense array."""
        arrays = tuple(map(self.get_block, sorted(self.sectors)))
        _concatenate = ar.get_lib_fn(self.backend, "concatenate")
        return _concatenate(arrays, axis=0)

    def allclose(self, other, **allclose_opts):
        return self._allclose_blockcommon(other, **allclose_opts)

    def test_allclose(self, other, **allclose_opts):
        """Like `allclose` but raises an AssertionError with details if not
        close."""
        return self._test_allclose_blockcommon(other, **allclose_opts)

    def __repr__(self):
        return "".join(
            [
                f"{self.__class__.__name__}(",
                f"total_size={self.size}, ",
                f"num_blocks={self.num_blocks})",
            ]
        )



import autoray as ar


def expm(x):
    """Matrix exponential of a AbelianArray."""
    if x.ndim != 2:
        raise NotImplementedError(
            "expm only implemented for 2D AbelianArrays,"
            f" got {x.ndim}D. Consider fusing first."
        )
    _expm = ar.get_lib_fn(x.backend, "scipy.linalg.expm")
    new = x.copy()
    new.apply_to_arrays(_expm)
    return new


# register non-scipy version
ar.register_function("symmray", "linalg.expm", expm)



import autoray as ar
import pytest

import symmray as sr


def invert_permutation(p):
    pinv = [0] * len(p)
    for i, v in enumerate(p):
        pinv[v] = i
    return tuple(pinv)


@pytest.mark.parametrize("symmetry", ("Z2", "U1"))
@pytest.mark.parametrize("d0", [3, 4])
@pytest.mark.parametrize("d1", [2, 5])
@pytest.mark.parametrize("f0", [False, True])
@pytest.mark.parametrize("f1", [False, True])
@pytest.mark.parametrize("c", [0, 1])
def test_qr_basics(symmetry, d0, d1, f0, f1, c):
    x = sr.utils.get_rand(
        symmetry,
        (d0, d1),
        duals=[f0, f1],
        charge=c,
        fermionic=True,
        oddpos="x",
    )
    x.check()
    q, r = sr.linalg.qr(x)
    q.check()
    r.check()
    assert (q @ r).allclose(x)


@pytest.mark.parametrize("symmetry", ("Z2", "U1"))
@pytest.mark.parametrize("d0", [3, 4])
@pytest.mark.parametrize("d1", [2, 5])
@pytest.mark.parametrize("f0", [False, True])
@pytest.mark.parametrize("f1", [False, True])
@pytest.mark.parametrize("c", [0, 1])
def test_svd_basics(symmetry, d0, d1, f0, f1, c):
    x = sr.utils.get_rand(
        symmetry,
        (d0, d1),
        duals=[f0, f1],
        charge=c,
        fermionic=True,
        subsizes="maximal",
        oddpos="x",
        seed=42,
    )
    x.check()
    u, s, vh = sr.linalg.svd(x)
    u.check()
    s.check()
    vh.check()
    assert (u @ ar.do("ldmul", s, vh)).allclose(x)


@pytest.mark.parametrize("symmetry", ("Z2", "U1", "Z2Z2", "U1U1"))
@pytest.mark.parametrize("seed", range(10))
def test_qr_roundtrip(symmetry, seed):
    rng = sr.utils.get_rng(seed)

    x = sr.utils.get_rand(
        symmetry=symmetry,
        shape=[4, 5, 6, 7, 8],
        fermionic=True,
        dist="uniform",
        seed=rng,
    )

    axes = tuple(rng.permutation(x.ndim))
    nleft = rng.integers(1, x.ndim - 1)

    axes_left = axes[:nleft]
    axes_right = axes[nleft:]
    order = (*axes_left, *axes_right)
    perm_back = invert_permutation(order)

    # fuse into matrix
    xf = x.fuse(axes_left, axes_right)

    # perform QR into matrix components
    q, r = ar.do("linalg.qr", xf)

    # reconstruct matrix
    xfr = q @ r

    # unfuse back into transpose tensor
    xrt = xfr.unfuse_all()

    # permute back to original order
    xr = xrt.transpose(perm_back)

    x.test_allclose(xr)


@pytest.mark.parametrize("symmetry", ("Z2", "U1", "U1U1", "Z2Z2"))
@pytest.mark.parametrize("seed", range(5))
def test_qr_with_expand_dims(symmetry, seed):
    x = sr.utils.get_rand(
        symmetry,
        [4, 5, 6],
        subsizes="maximal",
        seed=seed,
        fermionic=True,
    )
    y = x.reshape((1, 4 * 5 * 6))
    q, r = sr.linalg.qr(y)
    z = (q @ r).reshape((4, 5, 6))
    z.test_allclose(x)


@pytest.mark.parametrize("symmetry", ("Z2", "U1", "Z2Z2", "U1U1"))
@pytest.mark.parametrize("seed", range(10))
def test_svd_roundtrip(symmetry, seed):
    rng = sr.utils.get_rng(seed)

    x = sr.utils.get_rand(
        symmetry=symmetry,
        shape=[4, 5, 6, 7, 8],
        fermionic=True,
        dist="normal",
        seed=rng,
    )

    axes = tuple(rng.permutation(x.ndim))
    nleft = rng.integers(1, x.ndim - 1)

    axes_left = axes[:nleft]
    axes_right = axes[nleft:]
    order = (*axes_left, *axes_right)
    perm_back = invert_permutation(order)

    # fuse into matrix
    xf = x.fuse(axes_left, axes_right)

    # perform SVD into matrix components
    u, s, vh = ar.do("linalg.svd", xf)

    # reconstruct matrix
    xfr = u @ ar.do("multiply_diagonal", vh, s, axis=0)

    # unfuse back into transpose tensor
    xrt = xfr.unfuse_all()

    # permute back to original order
    xr = xrt.transpose(perm_back)

    x.test_allclose(xr)


@pytest.mark.parametrize("symmetry", ("Z2", "U1", "Z2Z2", "U1U1"))
@pytest.mark.parametrize("seed", range(10))
@pytest.mark.parametrize("absorb", [-1, 0, 1])
def test_svd_truncated_roundtrip(symmetry, seed, absorb):
    rng = sr.utils.get_rng(seed)

    x = sr.utils.get_rand(
        symmetry=symmetry,
        shape=[4, 5, 6, 7, 8],
        fermionic=True,
        dist="normal",
        seed=rng,
    )

    axes = tuple(rng.permutation(x.ndim))
    nleft = rng.integers(1, x.ndim - 1)

    axes_left = axes[:nleft]
    axes_right = axes[nleft:]
    order = (*axes_left, *axes_right)
    perm_back = invert_permutation(order)

    # fuse into matrix
    xf = x.fuse(axes_left, axes_right)

    # perform SVD into matrix components
    u, _, vh = ar.do("svd_truncated", xf, cutoff=1e-10, absorb=absorb)

    # reconstruct matrix
    xfr = u @ vh

    # unfuse back into transpose tensor
    xrt = xfr.unfuse_all()

    # permute back to original order
    xr = xrt.transpose(perm_back)

    x.test_allclose(xr)


@pytest.mark.parametrize("symmetry", ("Z2", "U1", "Z2Z2", "U1U1"))
@pytest.mark.parametrize("seed", range(20))
def test_svd_truncated_cutoff_max_bond(symmetry, seed):
    rng = sr.utils.get_rng(seed)

    x = sr.utils.get_rand(
        symmetry=symmetry,
        shape=[100, 100],
        fermionic=True,
        dist="uniform",
        subsizes=None,
        seed=rng,
    )

    # cutoff only
    _, s, _ = ar.do(
        "svd_truncated",
        x,
        cutoff=3e-2,
        absorb=None,
    )
    assert s.size < 80

    # max_bond only
    _, s, _ = ar.do(
        "svd_truncated",
        x,
        cutoff=0.0,
        max_bond=5,
        absorb=None,
    )
    assert s.size == 5

    # both
    _, s, _ = ar.do(
        "svd_truncated",
        x,
        cutoff=1e-2,
        max_bond=37,
        absorb=None,
    )
    assert s.size <= 37


@pytest.mark.parametrize("symm", ("Z2", "U1", "Z2Z2", "U1U1"))
@pytest.mark.parametrize("seed", range(10))
@pytest.mark.parametrize("dtype", ("float64", "complex128"))
def test_eigh_fermionic(symm, seed, dtype):
    d = 10
    i = sr.utils.rand_index(
        symm,
        d,
        subsizes=None,
        seed=seed,
    )
    x = sr.utils.get_rand(
        symm,
        shape=(i, i.conj()),
        fermionic=True,
        dtype=dtype,
        seed=seed,
    )
    # needs to be hermitian for eigh
    for sector, block in x.get_sector_block_pairs():
        x.set_block(sector, block + block.conj().T)
    el, ev = sr.linalg.eigh(x)
    # reconstruct the matrix
    y = sr.multiply_diagonal(ev, el, 1) @ ev.H
    y.test_allclose(x)


@pytest.mark.parametrize("symm", ("Z2", "U1", "Z2Z2", "U1U1"))
@pytest.mark.parametrize("seed", range(10))
@pytest.mark.parametrize("dtype", ("float64", "complex128"))
def test_eigh_truncated_fermionic(symm, seed, dtype):
    d = 10
    i = sr.utils.rand_index(
        symm,
        d,
        subsizes=None,
        seed=seed,
    )
    x = sr.utils.get_rand(
        symm,
        shape=(i, i.conj()),
        fermionic=True,
        dtype=dtype,
        seed=seed,
    )
    # needs to be hermitian for eigh
    for sector, block in x.get_sector_block_pairs():
        x.set_block(sector, block + block.conj().T)
    u, s, vh = sr.linalg.eigh_truncated(x, max_bond=-1, absorb=None)
    # reconstruct the matrix
    y = sr.multiply_diagonal(u, s, 1) @ vh
    y.test_allclose(x)

    z = u @ u.dagger_project_left() @ x @ vh.dagger_project_right() @ vh
    z.test_allclose(x)


@pytest.mark.parametrize("symm", ("Z2", "U1", "Z2Z2", "U1U1"))
@pytest.mark.parametrize("seed", range(10))
@pytest.mark.parametrize("dtype", ("float64", "complex128"))
def test_eigh_truncated_fermionic_fsued(symm, seed, dtype):
    import cotengra as ctg

    di = 4
    dj = 3
    i = sr.utils.rand_index(
        symm,
        di,
        subsizes=None,
        seed=seed,
    )
    j = sr.utils.rand_index(
        symm,
        dj,
        subsizes=None,
        seed=seed,
    )

    x = sr.utils.get_rand(
        symm,
        (i, j, i.conj(), j.conj()),
        fermionic=True,
        dtype=dtype,
    )
    x.randomize_phases(inplace=True)

    xf = x.fuse((0, 1), (2, 3))

    for sector, block in xf.get_sector_block_pairs():
        xf.set_block(sector, block + block.conj().T)

    # get unfused hermitian version
    x = xf.unfuse_all()

    # decompose
    u, s, vh = ar.do("eigh_truncated", xf, absorb=None)

    # reconstruct
    y = u.multiply_diagonal(s, axis=1) @ vh
    y.unfuse_all().test_allclose(x)

    # recostruct via identity resolved projectors
    udag = u.dagger_project_left().unfuse_all()
    vhdag = vh.dagger_project_right().unfuse_all()
    y = ctg.array_contract(
        arrays=[u.unfuse_all(), udag, x, vhdag, vh.unfuse_all()],
        inputs=[
            tuple("abc"),
            tuple("cde"),
            tuple("defg"),
            tuple("fgh"),
            tuple("hij"),
        ],
        output=tuple("abij"),
    )
    y.test_allclose(x)


@pytest.mark.parametrize("symm", ("Z2", "U1", "Z2Z2", "U1U1"))
@pytest.mark.parametrize("seed", range(10))
@pytest.mark.parametrize("b_charge", [False, True])
def test_solve_fermionic(symm, seed, b_charge):
    if symm in ("Z2", "U1"):
        b_charge = 1 if b_charge else 0
    elif symm in ("Z2Z2", "U1U1"):
        b_charge = (1, 0) if b_charge else (0, 1)

    i = sr.utils.rand_index(symm, 100, seed=seed)
    j = i.conj()
    A = sr.utils.get_rand(
        symm, shape=(i, j), fermionic=1, oddpos="a", seed=seed
    )
    b = sr.utils.get_rand(
        symm, shape=(i,), fermionic=1, charge=b_charge, oddpos="b", seed=seed
    )
    x = sr.linalg.solve(A, b)
    (A @ x).test_allclose(b)


@pytest.mark.parametrize("symm", ("Z2", "U1", "Z2Z2", "U1U1"))
@pytest.mark.parametrize("seed", range(10))
@pytest.mark.parametrize("b_charge", [False, True])
def test_solve_fermionic_2d(symm, seed, b_charge):
    if symm in ("Z2", "U1"):
        b_charge = 1 if b_charge else 0
    elif symm in ("Z2Z2", "U1U1"):
        b_charge = (1, 0) if b_charge else (0, 1)

    i = sr.utils.rand_index(symm, 100, seed=seed)
    j = i.conj()
    k = sr.utils.rand_index(symm, 20, seed=seed)
    A = sr.utils.get_rand(
        symm, shape=(i, j), fermionic=1, oddpos="a", seed=seed
    )
    b = sr.utils.get_rand(
        symm, shape=(i, k), fermionic=1, charge=b_charge, oddpos="b", seed=seed
    )
    x = sr.linalg.solve(A, b)
    (A @ x).test_allclose(b)


@pytest.mark.parametrize("symm", ("Z2", "U1", "Z2Z2", "U1U1"))
@pytest.mark.parametrize("seed", range(20))
def test_svd_projector_identity(symm, seed):
    import cotengra as ctg

    x = sr.utils.get_rand(
        symm,
        (4, 4, 5, 3),
        fermionic=True,
        seed=seed,
    )
    x.randomize_phases(inplace=True)
    xf = x.fuse((0, 1), (2, 3))

    u, s, vh = ar.do("linalg.svd", xf)
    udag = u.dagger_project_left()
    vhdag = vh.dagger_project_right()

    udag = udag.unfuse_all()
    vhdag = vhdag.unfuse_all()

    z = ctg.einsum(
        "abc,cde,defg,fgh,hij->abij",
        u.unfuse_all(),
        udag,
        x,
        vhdag,
        vh.unfuse_all(),
    )
    z.test_allclose(x)



"""Utility functions specifically for testing."""


def rand_valid_tensordot(
    symmetry,
    ndim_a=None,
    ndim_b=None,
    ncon=None,
    charge_a=None,
    charge_b=None,
    min_ndim=1,
    max_ndim=5,
    min_d=1,
    max_d=3,
    dimension_multiplier=1,
    subsizes=None,
    seed=None,
    **kwargs,
):
    """Generate two random symmetric arrays and valid contraction ``axes``.
    For testing purposes.

    Parameters
    ----------
    symmetry : str or Symmetry
        Symmetry to use.
    ndim_a : int or None
        Number of dimensions for the first array. If None, a random number is
        chosen between `min_ndim` and `max_ndim` (inclusive).
    ndim_b : int or None
        Number of dimensions for the second array. If None, a random number is
        chosen between `min_ndim` and `max_ndim` (inclusive).
    ncon : int or None
        Number of axes to contract. If None, a random number is chosen between
        0 and min(ndim_a, ndim_b) (inclusive).
    charge_a : charge_like or None
        Charge for the first array. If None, a random charge is chosen.
    charge_b : charge_like or None
        Charge for the second array. If None, a random charge is chosen.
    min_ndim : int
        Minimum number of dimensions for each array.
    max_ndim : int
        Maximum number of dimensions for each array.
    min_d : int
        Minimum dimension for each index. This is multiplied by
        `dimension_multiplier` if that is not None.
    max_d : int
        Maximum dimension for each index. This is multiplied by
        `dimension_multiplier` if that is not None.
    dimension_multiplier : int, optional
        Each index will have dimension `d * dimension_multiplier` where `d`
        is a random integer in [min_d, max_d]. Useful for generating flat
        arrays with exactly equally distributed charge sizes.
    subsizes : str or None
        Passed to `rand_index`. If "equal", all sectors will have equal size,
        required for flat arrays.
    seed : int or np.random.Generator or None
        Random seed or generator. If None, a new generator is created.
    **kwargs
        Additional arguments passed to `get_rand`.

    Returns
    -------
    a : AbelianArray
        First random array.
    b : AbelianArray
        Second random array.
    axes : tuple[tuple[int, ...], tuple[int, ...]]
        Axes to contract.
    """
    import numpy as np

    from .symmetries import get_symmetry
    from .utils import get_rand, get_rng, rand_index

    symmetry = get_symmetry(symmetry)
    rng = get_rng(seed)

    # dimensionality of each array
    if ndim_a is None:
        ndim_a = rng.integers(min_ndim, max_ndim + 1)
    if ndim_b is None:
        ndim_b = rng.integers(min_ndim, max_ndim + 1)

    # number of axes to contract
    if ncon is None:
        ncon = rng.integers(min(ndim_a, ndim_b) + 1)

    # which axes to contract
    axes_a = rng.choice(np.arange(ndim_a), size=ncon, replace=False)
    axes_b = rng.choice(np.arange(ndim_b), size=ncon, replace=False)
    axes = (tuple(map(int, axes_a)), tuple(map(int, axes_b)))

    # now construct compatible indices
    indices_a = [None] * ndim_a
    indices_b = [None] * ndim_b

    for axa, axb in zip(axes[0], axes[1]):
        d = int(dimension_multiplier * rng.integers(min_d, max_d + 1))
        ix = rand_index(symmetry, d, subsizes=subsizes, seed=rng)
        indices_a[axa] = ix.conj()
        indices_b[axb] = ix

    for i in range(ndim_a):
        if indices_a[i] is None:
            d = int(dimension_multiplier * rng.integers(min_d, max_d + 1))
            indices_a[i] = rand_index(symmetry, d, subsizes=subsizes, seed=rng)
    for i in range(ndim_b):
        if indices_b[i] is None:
            d = int(dimension_multiplier * rng.integers(min_d, max_d + 1))
            indices_b[i] = rand_index(symmetry, d, subsizes=subsizes, seed=rng)

    if charge_a is None:
        charge_a = symmetry.random_charge(rng)
    if charge_b is None:
        charge_b = symmetry.random_charge(rng)

    a = get_rand(
        symmetry,
        indices_a,
        charge=charge_a,
        seed=rng,
        oddpos="a",
        **kwargs,
    )
    b = get_rand(
        symmetry,
        indices_b,
        charge=charge_b,
        seed=rng,
        oddpos="b",
        **kwargs,
    )

    return a, b, axes



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "symmray"
copyright = "2022, Johnnie Gray"
author = "Johnnie Gray"

# The full version, including alpha/beta/rc tags
release = "0.0.1"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_nb",
    "sphinx.ext.napoleon",
]

nb_execution_mode = "off"
myst_heading_anchors = 4

# sphinx-autoapi
extensions.append("autoapi.extension")
autoapi_dirs = ["../symmray"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
# html_theme = 'sphinx_book_theme'

# TODO:
html_theme_options = {
    # "sidebar_hide_name": True,
    # "light_css_variables": {
    #     "color-brand-primary": "hsl(72, 75%, 40%)",
    #     "color-brand-content": "hsl(238, 50%, 60%)",
    # },
    # "dark_css_variables": {
    #     "color-brand-primary": "hsl(72, 75%, 60%)",
    #     "color-brand-content": "hsl(238, 75%, 70%)",
    # },
    # "light_logo": "logo-full.png",
    # "dark_logo": "logo-full.png",
}

# TODO: pygments_style = "lovelace"
# TODO: pygments_dark_style = "stata-dark"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# TODO:
# html_favicon = "_static/logo-favicon.ico"


def setup(app):
    app.add_css_file("my-styles.css")



import pytest

import symmray as sr

from .test_flat_abelian_array import get_zn_blocksparse_flat_compat


@pytest.mark.parametrize("symmetry", ["Z2", "Z3", "Z4"])
@pytest.mark.parametrize("d1", [6, 8])
@pytest.mark.parametrize("d2", [6, 8])
@pytest.mark.parametrize("charge", [0, 1])
@pytest.mark.parametrize("seed", [42, 34])
def test_flat_qr(symmetry, d1, d2, charge, seed):
    sx = get_zn_blocksparse_flat_compat(
        symmetry,
        shape=[d1, d2],
        charge=charge,
        seed=seed,
    )
    sQ, _, sR = sr.linalg.qr_stabilized(sx)
    fx = sx.to_flat()
    fx.check()
    fQ, _, fR = sr.linalg.qr_stabilized(fx)
    fQ.check()
    fR.check()

    assert fQ.charge == fx.charge

    assert fQ.to_blocksparse().allclose(sQ)
    assert fR.to_blocksparse().allclose(sR)

    fy = fQ @ fR
    fy.check()

    assert fy.charge == fx.charge

    assert fy.allclose(fx)
    assert fy.to_blocksparse().allclose(sx)


@pytest.mark.parametrize("symmetry", ("Z2",))
@pytest.mark.parametrize("seed", range(5))
def test_qr_with_expand_dims(symmetry, seed):
    pytest.xfail("Unfusing single dimensions not implemented yet")

    x = sr.utils.get_rand(
        symmetry,
        [4, 6, 6],
        seed=seed,
        flat=True,
        subsizes="maximal",
    )
    y = x.reshape((1, 4 * 6 * 6))
    q, r = sr.linalg.qr(y)
    xqr = q @ r
    z = xqr.reshape((4, 6, 6))
    z.test_allclose(x)


@pytest.mark.parametrize("symmetry", ["Z2", "Z3", "Z4"])
@pytest.mark.parametrize("d1", [6, 8])
@pytest.mark.parametrize("d2", [6, 8])
@pytest.mark.parametrize("charge", [0, 1])
@pytest.mark.parametrize("seed", [42, 34])
def test_flat_svd(symmetry, d1, d2, charge, seed):
    sx = get_zn_blocksparse_flat_compat(
        symmetry,
        shape=[d1, d2],
        charge=charge,
        seed=seed,
    )
    sU, sS, sVh = sr.linalg.svd(sx)
    fx = sx.to_flat()
    fx.check()
    fU, fS, fVh = sr.linalg.svd(fx)
    fU.check()
    fVh.check()

    assert fU.charge == fx.charge
    assert fS.to_blockvector().allclose(sS)

    assert fU.to_blocksparse().allclose(sU)
    assert fVh.to_blocksparse().allclose(sVh)

    fy = fU @ fVh.multiply_diagonal(fS, 0)
    fy.check()
    assert fy.charge == fx.charge
    assert fy.allclose(fx)
    assert fy.to_blocksparse().allclose(sx)

    fy = fU.multiply_diagonal(fS, 1) @ fVh
    fy.check()
    assert fy.charge == fx.charge
    assert fy.allclose(fx)
    assert fy.to_blocksparse().allclose(sx)


@pytest.mark.parametrize("symmetry", ["Z2", "Z3", "Z4"])
@pytest.mark.parametrize("d", [4, 8])
@pytest.mark.parametrize("seed", [42, 34])
@pytest.mark.parametrize("duals", [(0, 1), (1, 0)])
def test_eigh_flat(symmetry, d, seed, duals):
    sx = get_zn_blocksparse_flat_compat(
        symmetry,
        shape=[d, d],
        duals=duals,
        charge=0,
        seed=seed,
    )
    # need to make sure x is hermitian
    fx = sx.to_flat()

    sx = sx + sx.H
    fx = fx + fx.H

    fx.check()
    fevals, fvecs = sr.linalg.eigh(fx)
    fvecs.check()
    fevals.check()
    fy = fvecs.multiply_diagonal(fevals, 1) @ fvecs.H
    fy.check()
    assert fy.allclose(fx)



"""General interface for index objects."""


class Index:
    @property
    def dual(self) -> bool:
        """Whether the index flows 'outwards' / (+ve) / ket-like = ``False`` or
        'inwards' / (-ve) / bra-like= ``True``. The charge sign is given by
        ``(-1) ** dual``.
        """
        return self._dual

    @property
    def subinfo(self) -> "SubInfo":
        """Information about the subindices of this index and their extents if
        this index was formed from fusing.
        """
        return self._subinfo


class SubInfo:
    @property
    def indices(self) -> tuple[Index, ...]:
        """The indices that were fused to make this index."""
        return self._indices



"""Common interface functions for `symmray` array objects."""

import autoray as ar


def conj(x, **kwargs):
    """Conjugate a `symmray` array."""
    return x.conj(**kwargs)


def max(x):
    """Return the maximum value of a `symmray` array."""
    try:
        return x.max()
    except AttributeError:
        # called on non symmray array
        return ar.do("max", x)


def min(x):
    """Return the minimum value of a `symmray` array."""
    try:
        return x.min()
    except AttributeError:
        # called on non symmray array
        return ar.do("min", x)


def sum(x):
    """Return the sum of a `symmray` array."""
    try:
        return x.sum()
    except AttributeError:
        # called on non symmray array
        return ar.do("sum", x)


def all(x):
    """Check if all elements of a `symmray` array are true."""
    try:
        return x.all()
    except AttributeError:
        # called on non symmray array
        return ar.do("all", x)


def any(x):
    """Check if any elements of a `symmray` array are true."""
    try:
        return x.any()
    except AttributeError:
        # called on non symmray array
        return ar.do("any", x)


def isfinite(x):
    """Check if a `symmray` array contains only finite values."""
    try:
        return x.isfinite()
    except AttributeError:
        # called on non symmray array
        return ar.do("isfinite", x)


def abs(x):
    """Return the absolute value of a `symmray` array."""
    try:
        return x.abs()
    except AttributeError:
        # called on non symmray array
        return ar.do("abs", x)


def sqrt(x):
    """Return the square root of a `symmray` array."""
    try:
        return x.sqrt()
    except AttributeError:
        # called on non symmray array
        return ar.do("sqrt", x)


def log(x):
    """Return the natural logarithm of a `symmray` array."""
    try:
        return x.log()
    except AttributeError:
        # called on non symmray array
        return ar.do("log", x)


def log2(x):
    """Return the base-2 logarithm of a `symmray` array."""
    try:
        return x.log2()
    except AttributeError:
        # called on non symmray array
        return ar.do("log2", x)


def log10(x):
    """Return the base-10 logarithm of a `symmray` array."""
    try:
        return x.log10()
    except AttributeError:
        # called on non symmray array
        return ar.do("log10", x)


def real(x):
    """Return the real part of a `symmray` array."""
    try:
        return x.real
    except AttributeError:
        # called on non symmray array
        return ar.do("real", x)


def imag(x):
    """Return the imaginary part of a `symmray` array."""
    try:
        return x.imag
    except AttributeError:
        # called on non symmray array
        return ar.do("imag", x)


def clip(x, a_min, a_max):
    """Clip the values of a `symmray` array."""
    return x.clip(a_min, a_max)


def squeeze(x, axis=None):
    """Remove single-dimensional entries from the shape of a `symmray` array."""
    return x.squeeze(axis)


def expand_dims(x, axis):
    """Expand the shape of a `symmray` array."""
    return x.expand_dims(axis)


def reshape(a, newshape, **kwargs):
    """Reshape a `symmray` array, via fusing or unfusing."""
    return a.reshape(newshape, **kwargs)


def tensordot(a, b, axes=2, **kwargs):
    """Contract two `symmray` arrays along the specified axes.

    Parameters
    ----------
    a : AbelianArray or FermionicArray
        First array to contract.
    b : AbelianArray or FermionicArray
        Second array to contract, with same type as `a`.
    axes : int or tuple of int, optional
        If an integer, the number of axes to contract. If a tuple, the axes
        to contract. Default is 2.
    """
    try:
        return a.tensordot(b, axes, **kwargs)
    except AttributeError:
        if getattr(a, "ndim", 0) == 0:
            # likely called as effective scalar multiplication of block array
            return a * b
        else:
            raise TypeError(f"Expected SymmrayCommon, got {type(a).__name__}.")


def einsum(eq, x):
    """Perform an Einstein summation on a `symmray` array."""
    return x.einsum(eq)


def transpose(a, axes=None, **kwargs):
    """Transpose a `symmray` array."""
    return a.transpose(axes, **kwargs)


def trace(a):
    """Return the trace of a `symmray` array."""
    return a.trace()


def take(a, indices, axis, **kwargs):
    """Take elements from a `symmray` array along an axis."""
    return a.take(indices, axis, **kwargs)


# non-standard 'composed' functions


def multiply_diagonal(x, v, axis):
    """Multiply a `symmray` array by a vector as if contracting a diagonal
    matrix into one axis.
    """
    return x.multiply_diagonal(v, axis)


ar.register_function("symmray", "multiply_diagonal", multiply_diagonal)


def ldmul(v, x):
    """Left multiply a `symmray` matrix `x` by a vector `v`."""
    return x.ldmul(v)


ar.register_function("symmray", "ldmul", ldmul)


def rdmul(x, v):
    """Right multiply a `symmray` matrix `x` by a vector `v`."""
    return x.rdmul(v)


ar.register_function("symmray", "rdmul", rdmul)


def lddiv(v, x):
    """Left divide a `symmray` matrix `x` by a vector `v`."""
    return x.lddiv(v)


ar.register_function("symmray", "lddiv", lddiv)


def rddiv(x, v):
    """Right divide a `symmray` matrix `x` by a vector `v`."""
    return x.rddiv(v)


ar.register_function("symmray", "rddiv", rddiv)


def align_axes(x, y, axes):
    """ """
    return x.align_axes(y, axes)


ar.register_function("symmray", "align_axes", align_axes)


def fuse(x, *axes_groups):
    """Fuse multiple axes of a `symmray` array."""
    return x.fuse(*axes_groups)


ar.register_function("symmray", "fuse", fuse)



"""Helper functions for building local fermionic operators, with 'internal'
signs pre-computed.
"""

import itertools
import numbers

import autoray as ar


def labels_lt(labela, labelb):
    try:
        return labela < labelb
    except TypeError:
        # allow certain mixed types to be compared, the logic here is:
        # 1. group all integer labels first
        # 2. then plain strings
        # 3. then tuples by length
        # XXX: some different length tuples can be compared, might revisit
        if isinstance(labela, numbers.Integral):
            la = -2
        elif isinstance(labela, str):
            la = -1
        else:
            la = len(labela)

        if isinstance(labelb, numbers.Integral):
            lb = -2
        elif isinstance(labelb, str):
            lb = -1
        else:
            lb = len(labelb)

        return la < lb


class FermionicOperator:
    """Simple class to represent a fermionic operator with a label and a
    dual flag.
    """

    __slots__ = ("_label", "_dual")

    def __init__(self, label, dual=False):
        if label == ():
            raise ValueError("Label cannot be an empty tuple.")
        self._label = label
        self._dual = dual

    @property
    def label(self):
        return self._label

    @property
    def dual(self):
        return self._dual

    @property
    def dag(self):
        return FermionicOperator(self._label, not self._dual)

    def __eq__(self, other):
        return (self._dual, self._label) == (other._dual, other._label)

    def __lt__(self, other):
        if self.dual:
            if other.dual:
                # dual operator are reflected
                return labels_lt(other.label, self.label)
            else:
                # creation left of annihilation
                return True
        else:
            if other.dual:
                return False
            else:
                return labels_lt(self.label, other.label)

    def __repr__(self):
        return f"{self._label}{'+' if self._dual else '-'}"


def _dagger_basis(basis):
    """Return the conjugate basis of a given basis."""
    return tuple(tuple(op.dag for op in reversed(x)) for x in basis)


def _ensure_fermionic_operator(op):
    """Possibly convert a tuple of (label, symbol) to a FermionicOperator."""
    if isinstance(op, FermionicOperator):
        return op
    label, symbol = op
    return FermionicOperator(label, {"+": True, "-": False}[symbol])


def _parse_terms(terms):
    """Allow ops to be specified as tuples of (site, symbol) in addition to
    FermionicOperator instances. E.g. ``(-t, [('a', '+'), ('b', '-')])``.
    """
    return tuple(
        (coeff, tuple(map(_ensure_fermionic_operator, term)))
        for coeff, term in terms
    )


def _parse_bases(bases):
    """Allow ops to be specified as tuples of (site, symbol) in addition to
    FermionicOperator instances.
    """
    return tuple(
        tuple(tuple(map(_ensure_fermionic_operator, ops)) for ops in basis)
        for basis in bases
    )


def build_local_fermionic_elements(terms, bases):
    """Compute the elements of a local fermionic operator in a given tensor
    basis, including 'internal' signs.

    Parameters
    ----------
    terms : tuple[tuple[float, tuple[FermionicOperator, ...]]]
        The terms in the operator, each a tuple of a coefficient and a tuple
        of FermionicOperator or tuple[label, op] instances.
    bases : tuple[tuple[tuple[FermionicOperator]]]
        The tensor bases to compute the operator elements in. Each basis is a
        sequence of multiple FermionicOperator instancess acting on the vacuum.

    Returns
    -------
    list[tuple[tuple[int], float]]
        A list of tuples of tensor indices and the corresponding tensor
        element, including phases.

    Examples
    --------
    Compute the elements of a local fermionic operator in a tensor basis::

        a, b = map(FermionicOperator, "ab")
        basis_a = [(), (a.dag,)]
        basis_b = [(), (b.dag,)]
        bases = (basis_a, basis_b)

        t = 1.0
        U = 8.0

        terms = (
            (-t, (a.dag, b)),
            (-t, (b.dag, a)),
            (U, (a.dag, a, b.dag, b)),
        )

        build_local_fermionic_elements(terms, bases)
        # {(0, 1, 1, 0): -1.0, (1, 0, 0, 1): -1.0, (1, 1, 1, 1): -8.0}

    """
    terms = _parse_terms(terms)
    bases = _parse_bases(bases)

    # construct |i>|j>|k> along with linear tensor indices
    enum_right_bases = [tuple(enumerate(x)) for x in bases]
    # and <i'|<j'|<k'|  n.b. not <k'|<j'|<i'|
    enum_left_bases = [tuple(enumerate(_dagger_basis(x))) for x in bases]

    all_locations = itertools.product(
        itertools.product(*enum_left_bases),
        itertools.product(*enum_right_bases),
    )

    entries = {}
    for l, r in all_locations:
        # get tensor index, and left and right
        # basis operators it corresponds to
        left_indices, left_basis_ops = zip(*l)
        right_indices, right_basis_ops = zip(*r)

        # flatten into single lists of operators
        left_basis_ops = [op for ops in left_basis_ops for op in ops]
        right_basis_ops = [op for ops in right_basis_ops for op in ops]

        # process all terms at this tensor index
        for coeff, term in terms:
            if coeff == 0.0:
                continue

            # sandwich terms between basis operators
            element = [*left_basis_ops, *term, *right_basis_ops]

            # phased sort by labels
            phase = 1
            any_moves = True
            while any_moves:
                any_moves = False
                for k in range(len(element) - 1):
                    # check each pair of neighboring operators
                    opl = element[k]
                    opr = element[k + 1]
                    if opl.label > opr.label:
                        # swap if not grouped contiguously yet
                        element[k] = opr
                        element[k + 1] = opl
                        # flip phase from fermionic anticommutation
                        phase = -phase
                        any_moves = True

            # group by label
            groups = {}
            for x in element:
                groups.setdefault(x.label, []).append(x)

            # all groups must be like <0|(- + - + ... - +)|0> to not vanish
            nonvanishing = all(
                (
                    (len(group) % 2 == 0)
                    and all(not op.dual for op in group[::2])
                    and all(op.dual for op in group[1::2])
                )
                for group in groups.values()
            )

            if nonvanishing:
                # a non-vanishing element
                index = (*left_indices, *right_indices)
                entries[index] = entries.get(index, 0.0) + phase * coeff

    return entries


def build_local_fermionic_dense(terms, bases, like="numpy"):
    hij = ar.do("zeros", tuple(len(b) for b in bases) * 2, like=like)

    for idx, val in build_local_fermionic_elements(terms, bases).items():
        hij[idx] += val

    return hij


def build_local_fermionic_array(
    terms,
    bases,
    symmetry,
    index_maps,
    like="numpy",
):
    """Compute a local fermionic operator as a `FermionicArray`.

    Parameters
    ----------
    terms : tuple[tuple[float, tuple[FermionicOperator, ...]]]
        The terms in the operator, each a tuple of a coefficient and a tuple
        of FermionicOperator instances.
    bases : tuple[tuple[tuple[FermionicOperator]]]
        The tensor bases to compute the operator elements in. Each basis is a
        sequence of multiple FermionicOperator instances acting on the vacuum.
    symmetry : str
        The symmetry of the model. Either "Z2", "U1", "Z2Z2" or "U1U1".
    index_maps : Sequence[Sequence[hashable]]
        For each basis, the sequence mapping linear index to charge sector.

    Returns
    -------
    FermionicArray
        The local operator in fermionic array form.
    """
    from .utils import from_dense

    dense = build_local_fermionic_dense(terms, bases, like=like)
    duals = [False] * len(bases) + [True] * len(bases)

    return from_dense(
        dense,
        duals=duals,
        symmetry=symmetry,
        fermionic=True,
        index_maps=index_maps * 2,
    )


def get_spinless_charge_indexmap(symmetry):
    """Get a mapping of linear index to charge sector for a spinless
    fermion model.

    Parameters
    ----------
    symmetry : str
        The symmetry of the model. Either "Z2" or "U1".

    Returns
    -------
    list[hashable]
    """
    if symmetry in ("Z2", "U1"):
        return [0, 1]
    else:
        raise ValueError(f"Invalid symmetry: {symmetry}")


def get_spinful_charge_indexmap(symmetry):
    """Get a mapping of linear index to charge sector for a spinful
    fermion model.

    Parameters
    ----------
    symmetry : str
        The symmetry of the model. Either "Z2", "U1", "Z2Z2", or "U1U1".

    Returns
    -------
    list[hashable]
    """
    if symmetry == "Z2":
        return [0, 1, 1, 0]
    elif symmetry == "U1":
        return [0, 1, 1, 2]
    elif symmetry in ("Z2Z2", "U1U1"):
        return [(0, 0), (0, 1), (1, 0), (1, 1)]
    else:
        raise ValueError(f"Invalid symmetry: {symmetry}")


def fermi_hubbard_spinless_local_array(
    symmetry,
    t=1.0,
    V=8.0,
    mu=0.0,
    coordinations=(1, 1),
    like="numpy",
):
    """Construct the fermionic local tensor for the spinless Fermi-Hubbard
    model. The indices are ordered as (a, b, a', b').

    Parameters
    ----------
    symmetry : str
        The symmetry of the model. Either "Z2" or "U1".
    t : float, optional
        The hopping parameter, by default 1.0.
    V : float, optional
        The nearest-neighbor interaction parameter, by default 8.0.
    mu : float or (float, float), optional
        The chemical potential, by default 0.0. If a tuple, then the chemical
        potential is different for each site.
    coordinations : tuple[int, int], optional
        The coordinations of the sites, by default (1, 1). If applying this
        local operator to every edge in a graph, then the single site
        contributions can be properly accounted for if the coordinations are
        provided.
    like : str, optional
        The backend to use, by default "numpy".

    Returns
    -------
    array
        The local operator in fermionic array form.
    """
    a, b = map(FermionicOperator, "ab")

    try:
        mua, mub = mu
    except TypeError:
        mua = mub = mu

    terms = (
        # hopping
        (-t, (a.dag, b)),
        (-t, (b.dag, a)),
        # nearest-neighbor interaction
        (V, (a.dag, a, b.dag, b)),
        # chemical potential
        # mu is single site and will be overcounted without coordinations
        (-mua / coordinations[0], (a.dag, a)),
        (-mub / coordinations[1], (b.dag, b)),
    )

    basis_a = ((), (a.dag,))
    basis_b = ((), (b.dag,))
    bases = (basis_a, basis_b)
    indexmap = get_spinless_charge_indexmap(symmetry)

    return build_local_fermionic_array(
        terms,
        bases,
        symmetry,
        index_maps=[indexmap, indexmap],
        like=like,
    )


def fermi_hubbard_local_array(
    symmetry,
    t=1.0,
    U=8.0,
    mu=0.0,
    coordinations=(1, 1),
    like="numpy",
):
    """Construct the fermionic local tensor for the Fermi-Hubbard model. The
    indices are ordered as (a, b, a', b'), with the local basis like
    (|00>, ad+|00>, au+|00>, au+ad+|00>) for site a with up (au) and down (ad)
    spin respectively and similar for site b.

    Parameters
    ----------
    symmetry : str
        The symmetry of the model. Either "Z2", "U1", "Z2Z2", or "U1U1".
    t : float, optional
        The hopping parameter, by default 1.0.
    U : float or (float, float), optional
        The interaction parameter, by default 8.0. If a tuple, then the
        interaction parameter is different for each site.
    mu : float or (float, float), optional
        The chemical potential, by default 0.0. If a tuple, then the chemical
        potential is different for each site.
    coordinations : tuple[int, int], optional
        The coordinations of the sites, by default (1, 1). If applying this
        local operator to every edge in a graph, then the single site
        contributions can be properly accounted for if the coordinations are
        provided.
    like : str, optional
        The backend to use, by default "numpy".

    Returns
    -------
    array : FermionicArray
        The local operator in fermionic array form.
    """
    au = FermionicOperator("au")
    ad = FermionicOperator("ad")
    bu = FermionicOperator("bu")
    bd = FermionicOperator("bd")

    try:
        Ua, Ub = U
    except TypeError:
        Ua = Ub = U

    try:
        mua, mub = mu
    except TypeError:
        mua = mub = mu

    terms = [
        (-t, (au.dag, bu)),
        (-t, (bu.dag, au)),
        (-t, (ad.dag, bd)),
        (-t, (bd.dag, ad)),
        # U, mu are single site and will be overcounted without coordinations
        (Ua / coordinations[0], (au.dag, au, ad.dag, ad)),
        (Ub / coordinations[1], (bu.dag, bu, bd.dag, bd)),
        (-mua / coordinations[0], (au.dag, au)),
        (-mua / coordinations[0], (ad.dag, ad)),
        (-mub / coordinations[1], (bu.dag, bu)),
        (-mub / coordinations[1], (bd.dag, bd)),
    ]

    basis_a = ((), (ad.dag,), (au.dag,), (au.dag, ad.dag))
    basis_b = ((), (bd.dag,), (bu.dag,), (bu.dag, bd.dag))
    bases = [basis_a, basis_b]
    indexmap = get_spinful_charge_indexmap(symmetry)

    return build_local_fermionic_array(
        terms,
        bases,
        symmetry,
        index_maps=[indexmap, indexmap],
        like=like,
    )


def fermi_number_operator_spinless_local_array(symmetry, like="numpy"):
    """Construct the fermionic number operator for the spinless Fermi-Hubbard
    model. The indices are ordered as (a, a'). The local basis is like
    (|0>, a+|0>) for single site a.

    Parameters
    ----------
    symmetry : str
        The symmetry of the model. Either "Z2" or "U1".
    like : str, optional

    Returns
    -------
    array : FermionicArray
        The local operator in fermionic array form.
    """
    a = FermionicOperator("a")

    terms = [(1, (a.dag, a))]
    bases = [((), (a.dag,))]
    indexmap = get_spinless_charge_indexmap(symmetry)

    return build_local_fermionic_array(
        terms,
        bases,
        symmetry,
        index_maps=[indexmap],
        like=like,
    )


def fermi_number_operator_spinful_local_array(symmetry, like="numpy"):
    """Construct the fermionic number operator for the Fermi-Hubbard model. The
    indices are ordered as (a, a'), with the local basis like
    (|00>, ad+|00>, au+|00>, au+ad+|00>) for site a with up (au) and down (ad)
    spin respectively for single site `a`.

    Parameters
    ----------
    symmetry : str
        The symmetry of the model. Either "Z2", "U1", "Z2Z2", or "U1U1".
    like : str, optional
        The backend to use, by default "numpy".

    Returns
    -------
    array : FermionicArray
        The local operator in fermionic array form.
    """
    au = FermionicOperator("au")
    ad = FermionicOperator("ad")

    # nup + ndown
    terms = [(1, (au.dag, au)), (1, (ad.dag, ad))]
    # |00>, |01>, |10>, |11>
    bases = [((), (ad.dag,), (au.dag,), (au.dag, ad.dag))]
    indexmap = get_spinful_charge_indexmap(symmetry)

    return build_local_fermionic_array(
        terms,
        bases,
        symmetry,
        index_maps=[indexmap],
        like=like,
    )


def fermi_number_up_local_array(symmetry, like="numpy"):
    """Construct the 'up' fermionic number operator for the Fermi-Hubbard
    model. The indices are ordered as (a, a'), with the local basis like
    (|00>, ad+|00>, au+|00>, au+ad+|00>) for site a with up (au) and down (ad)
    spin respectively for single site `a`.

    Parameters
    ----------
    symmetry : str
        The symmetry of the model. Either "Z2", "U1", "Z2Z2", or "U1U1".
    like : str, optional
        The backend to use, by default "numpy".

    Returns
    -------
    array : FermionicArray
        The local operator in fermionic array form.
    """
    au = FermionicOperator("au")
    ad = FermionicOperator("ad")

    # nup
    terms = [(1, (au.dag, au))]
    # |00>, |01>, |10>, |11>
    bases = [((), (ad.dag,), (au.dag,), (au.dag, ad.dag))]
    indexmap = get_spinful_charge_indexmap(symmetry)

    return build_local_fermionic_array(
        terms,
        bases,
        symmetry,
        index_maps=[indexmap],
        like=like,
    )


def fermi_number_down_local_array(symmetry, like="numpy"):
    """Construct the 'down' fermionic number operator for the Fermi-Hubbard
    model. The indices are ordered as (a, a'), with the local basis like
    (|00>, ad+|00>, au+|00>, au+ad+|00>) for site a with up (au) and down (ad)
    spin respectively for single site `a`.

    Parameters
    ----------
    symmetry : str
        The symmetry of the model. Either "Z2", "U1", "Z2Z2", or "U1U1".
    like : str, optional
        The backend to use, by default "numpy".

    Returns
    -------
    array : FermionicArray
        The local operator in fermionic array form.
    """
    au = FermionicOperator("au")
    ad = FermionicOperator("ad")

    # nup
    terms = [(1, (ad.dag, ad))]
    # |00>, |01>, |10>, |11>
    bases = [((), (ad.dag,), (au.dag,), (au.dag, ad.dag))]
    indexmap = get_spinful_charge_indexmap(symmetry)

    return build_local_fermionic_array(
        terms,
        bases,
        symmetry,
        index_maps=[indexmap],
        like=like,
    )


def fermi_spin_operator_local_array(symmetry, like="numpy"):
    """Construct the fermionic spin operator for the Fermi-Hubbard model. The
    indices are ordered as (a, a'), with the local basis like
    (|00>, ad+|00>, au+|00>, au+ad+|00>) for site a with up (au) and down (ad)
    spin respectively for single site `a`.

    Parameters
    ----------
    symmetry : str
        The symmetry of the model. Either "Z2", "U1", "Z2Z2", or "U1U1".
    like : str, optional
        The backend to use, by default "numpy".

    Returns
    -------
    array : FermionicArray
        The local operator in fermionic array form.
    """
    au = FermionicOperator("au")
    ad = FermionicOperator("ad")

    # S^z = 1/2 (nup - ndown)
    terms = [(0.5, (au.dag, au)), (-0.5, (ad.dag, ad))]
    # |00>, |01>, |10>, |11>
    bases = [((), (ad.dag,), (au.dag,), (au.dag, ad.dag))]
    indexmap = get_spinful_charge_indexmap(symmetry)

    return build_local_fermionic_array(
        terms,
        bases,
        symmetry,
        index_maps=[indexmap],
        like=like,
    )



import functools
import hashlib
import numbers
import os
import pickle

# a simple flag for enabling rigorous checks in many places
DEBUG = bool(os.environ.get("SYMMRAY_DEBUG", "0").upper() in ("1", "TRUE"))


def set_debug(debug):
    global DEBUG
    DEBUG = debug


def lazyabstractmethod(method):
    """Mark a method as one that must be implemented in a subclass, but only
    enforce this when the method is called. This can be used as a decorator (if
    you want to demonstrate the call signature) or by directly assigning the
    result to a method name.
    """

    if callable(method):
        name = method.__name__
    else:
        name = str(method)

    def raising_method(self, *args, **kwargs):
        raise NotImplementedError(
            f"`{name}` must be implemented in "
            f"subclass `{self.__class__.__name__}`"
        )

    return raising_method


def hasher(k):
    return hashlib.sha1(pickle.dumps(k)).hexdigest()


class RandomStateTranslated:
    """Simple wrapper to make `numpy.random.RandomState` have the same
    interface as `numpy.random.Generator`."""

    def __init__(self, rng):
        self.rng = rng

    def integers(self, *args, **kwargs):
        return self.rng.randint(*args, **kwargs)

    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            x = getattr(self.rng, name)
            super().__getattribute__("__dict__")[name] = x
            return x


def get_rng(seed=None):
    import numpy as np

    # RandomStateTranslated is useful for determinism across numpy versions
    if isinstance(seed, RandomStateTranslated):
        return seed
    elif isinstance(seed, np.random.RandomState):
        return RandomStateTranslated(seed)
    else:
        return np.random.default_rng(seed)


def get_random_fill_fn(
    seed=None,
    dist="normal",
    dtype="float64",
    scale=1.0,
    loc=0.0,
):
    """Get a function that produces numpy arrays of random numbers with the
    specified distribution, dtype, loc, and scale.

    Parameters
    ----------
    seed : None, int, or numpy.random.Generator, optional
        The seed for the random number generator.
    dist : str, optional
        The distribution of the random numbers. Can be "normal" or "uniform" or
        any other distribution supported by numpy.
    dtype : str, optional
        The data type of the random numbers. If "complex", the real and
        imaginary parts are generated separately and added.
    scale : float, optional
        A multiplicative factor to the distribution.
    loc : float, optional
        An additive offset to the distribution.

    Returns
    -------
    callable
        A function with signature `fill_fn(shape) -> numpy.ndarray`.
    """
    rng = get_rng(seed)

    def fill_fn(shape):
        x = getattr(rng, dist)(size=shape)
        if "complex" in dtype:
            x = x + 1j * getattr(rng, dist)(size=shape)
        if scale != 1.0:
            x *= scale
        if loc != 0.0:
            x += loc
        if x.dtype != dtype:
            x = x.astype(dtype)

        return x

    return fill_fn


def rand_partition(d, n, seed=None):
    """Randomly partition `d` into `n` sizes each of size at least 1."""
    if d == n:
        return [1] * n

    rng = get_rng(seed)

    if n == 2:
        # cut in two
        s = int(rng.integers(1, d))
        return [s, d - s]

    # cut into 3 or more
    splits = (
        0,
        *sorted(rng.choice(range(1, d - 1), size=n - 1, replace=False)),
        d,
    )
    return [int(splits[i + 1] - splits[i]) for i in range(n)]


@functools.cache
def get_zn_charges(ncharge, order):
    """Get a list of ``ncharge`` distinct ZN charges that are as close to
    0 or |G| as possible, with a slight bias towards positive charges.

    Parameters
    ----------
    ncharge : int
        The number of distinct charges to get.
    order : int
        The order (i.e. size, N) of the cyclic group ZN.

    Returns
    -------
    charges : list[int]
        A list of `ncharge <= order` distinct ZN charges.
    """
    if ncharge >= order:
        return list(range(order))

    # get charges 'close to' zero / |G|
    return sorted(
        c % order for c in range(-ncharge // 2 + 1, +ncharge // 2 + 1)
    )


def rand_zn_index(
    d,
    order,
    dual="random",
    subsizes="random",
    seed=None,
):
    """Generate a random ZN index with the given dimension and order.

    Parameters
    ----------
    d : int, dict[hashable, int] or sequence[(hashable, int), ...]
        The total size of the index. If a dict, an explicit chargemap. If a
        sequence, a linearmap.
    order : int
        The order (i.e. size, N) of the cyclic group ZN.
    dual : bool or "random", optional
        The dualness of the index. If "random", it is randomly chosen.
    subsizes : "random", "equal", "maximal", "minimal", or tuple[int], optional
        The sizes of the charge sectors. The choices are as follows:

        - "random": the charges and sizes are randomly determined.
        - "equal": a fixed number of charges 'close to' zero charge are chosen,
          all with equal size (up to remainders).
        - "maximal": as many charges as possible are chosen, each with size 1
          (or more if the total number of charges is less than the total size).
        - "minimal": only the zero charge sector is chosen, with full size.
        - tuple: the sizes of the charge sectors, a matching number of charges
          are chosen automatically, in sequence 'closest to zero'.

    seed : None, int, or numpy.random.Generator, optional
        The seed for the random number generator.

    Returns
    -------
    BlockIndex
    """
    import symmray as sr

    rng = get_rng(seed)

    if (dual is None) or (dual == "random"):
        dual = rng.choice([False, True])

    # convert from possible numpy.int etc.
    d = int(d)

    if d == 1:
        # can only have one charge sector
        if (subsizes is None) or (subsizes == "random"):
            charge = int(rng.integers(order))
        else:
            charge = 0
        return sr.BlockIndex(chargemap={charge: 1}, dual=dual)

    if (subsizes is None) or (subsizes == "random"):
        # randomly distributed over all
        ncharge = min(d, order)
        subsizes = rand_partition(d, ncharge, seed=rng)

    elif subsizes in ("equal", "maximal"):
        # round-robin == maximal spread
        ncharge = min(d, order)
        subsizes = [
            d // ncharge + int(i < d % ncharge) for i in range(ncharge)
        ]

    elif subsizes == "minimal":
        # all in zero charge sector
        ncharge = 1
        subsizes = [d]

    else:
        # sizes given explicitly
        ncharge = len(subsizes)

    charges = get_zn_charges(ncharge, order)
    chargemap = dict(zip(charges, subsizes))

    return sr.BlockIndex(chargemap=chargemap, dual=dual)


rand_z2_index = functools.partial(rand_zn_index, order=2)


def rand_z2z2_index(
    d,
    dual="random",
    subsizes="random",
    seed=None,
):
    import symmray as sr

    rng = get_rng(seed)

    if (dual is None) or (dual == "random"):
        dual = rng.choice([False, True])

    possible = [(0, 0), (0, 1), (1, 0), (1, 1)]

    if (subsizes is None) or (subsizes == "random"):
        # randomly distributed
        if d < 4:
            charges = rng.choice(possible, size=d, replace=False)
            chargemap = {tuple(c): 1 for c in charges}
        else:
            subsizes = rand_partition(d, 4, seed=rng)
            chargemap = dict(zip(possible, subsizes))

    elif subsizes in ("equal", "maximal"):
        # round-robin distribution
        ncharge = min(d, 4)
        charges = possible[:ncharge]
        chargemap = {c: d // 4 + (i < d % 4) for i, c in enumerate(charges)}

    elif subsizes == "minimal":
        # all in zero charge sector
        chargemap = {(0, 0): d}

    else:
        # sizes given explicitly
        ncharge = len(subsizes)
        chargemap = dict(zip(possible, subsizes))

    return sr.BlockIndex(chargemap=chargemap, dual=dual)


def get_u1_charges(ncharge):
    """Get a list of ``ncharge`` distinct U1 charges that are as close to the
    origin as possible, with a slight bias towards positive charges.
    """
    charges = list(range(-ncharge // 2 + 1, +ncharge // 2 + 1))
    charges.sort(key=lambda x: (abs(x), -x))
    return tuple(charges[:ncharge])


def rand_u1_index(
    d,
    dual="random",
    subsizes="random",
    seed=None,
):
    """Generate a random U1 index with the given dimension.

    Parameters
    ----------
    d : int
        The total size of the index.
    dual : bool or "random", optional
        The dualness of the index. If "random", it is randomly chosen.
    subsizes : "random", "equal", "maximal", "minimal", or tuple[int], optional
        The sizes of the charge sectors. The choices are as follows:

        - "random": the charges and sizes are randomly determined.
        - "equal": a fixed number of charges 'close to' zero charge are chosen,
          all with equal size (up to remainders).
        - "maximal": as many charges as possible are chosen, each with size 1
          (or more if the total number of charges is less than the total size).
        - "minimal": only the zero charge sector is chosen, with full size.
        - tuple: the sizes of the charge sectors, a matching number of charges
          are chosen automatically, in sequence 'closest to zero'.

    seed : None, int, or numpy.random.Generator, optional
        The seed for the random number generator.

    Returns
    -------
    BlockIndex
    """
    import symmray as sr

    rng = get_rng(seed)

    if (dual is None) or (dual == "random"):
        dual = rng.choice([False, True])

    if (subsizes is None) or (subsizes == "random"):
        # a random number and distribution of charges
        ncharge = rng.integers(1, d + 1)
        subsizes = rand_partition(d, ncharge, seed=rng)

    elif subsizes == "equal":
        # 3 approx equal charges around the origin
        ncharge = min(d, 3)
        subsizes = [
            d // ncharge + int(i < d % ncharge) for i in range(ncharge)
        ]

    elif subsizes == "maximal":
        # maximal spread of charges each with size 1
        ncharge = d
        subsizes = [1 for _ in range(ncharge)]

    elif subsizes == "minimal":
        # minimal spread of charges each with size 1
        ncharge = 1
        subsizes = [d]

    else:
        # only sizes given explicitly
        ncharge = len(subsizes)

    charges = get_u1_charges(ncharge)
    chargemap = dict(zip(charges, subsizes))

    return sr.BlockIndex(chargemap=chargemap, dual=dual)


def get_u1u1_charges(ncharge):
    """Get a list of ``ncharge`` distinct U1U1 charges that are as close to
    the origin as possible.
    """
    import itertools

    k = int(ncharge**0.5)
    krange = tuple(range(-k + 1, k + 1))
    charges = []
    for i, j in itertools.product(krange, repeat=2):
        charges.append((i, j))

    charges.sort(
        # choose spherical distribution with positive tie breaking bias
        key=lambda xy: (xy[0] ** 2 + xy[1] ** 2, -xy[0] - xy[1])
        # # diamond distribution:
        # key=lambda xy: (abs(xy[0]) + abs(xy[1]), -xy[0], -xy[1])
    )

    return tuple(charges[:ncharge])


def rand_u1u1_index(
    d,
    dual="random",
    subsizes="random",
    seed=None,
):
    import symmray as sr

    rng = get_rng(seed)

    if (dual is None) or (dual == "random"):
        dual = rng.choice([False, True])

    if (subsizes is None) or (subsizes == "random"):
        # a random number and distribution of charges
        ncharge = rng.integers(1, d + 1)
        subsizes = rand_partition(d, ncharge, seed=rng)

    elif subsizes == "equal":
        # 9 approx equal charges around the origin
        ncharge = min(d, 9)
        subsizes = [
            d // ncharge + int(i < d % ncharge) for i in range(ncharge)
        ]

    elif subsizes == "maximal":
        # maximal spread of charges each with size 1
        ncharge = d
        subsizes = [1 for _ in range(ncharge)]

    elif subsizes == "minimal":
        # all in zero charge sector
        ncharge = 1
        subsizes = [d]

    else:
        # only sizes given explicitly
        ncharge = len(subsizes)

    charges = get_u1u1_charges(ncharge)
    chargemap = dict(zip(charges, subsizes))

    return sr.BlockIndex(chargemap=chargemap, dual=dual)


def rand_index(
    symmetry,
    d,
    dual="random",
    subsizes="random",
    seed=None,
):
    """Get a random index with the given symmetry.

    Parameters
    ----------
    symmetry : str or Symmetry
        The symmetry of the index.
    d : int, dict, sequence, or BlockIndex
        The total size of the index. If a dict, an explicit chargemap. If a
        sequence an explicit linearmap. If a BlockIndex, it is returned
        unchanged.
    dual : bool or "random", optional
        The dualness of the index. If "random", it is randomly chosen.
    subsizes : "random", "equal", "maximal", "minimal", or tuple[int], optional
        The sizes of the charge sectors. The choices are as follows:

        - "random": the charges and sizes are randomly determined.
        - "equal": a fixed number of charges 'close to' zero charge are chosen,
          all with equal size (up to remainders).
        - "maximal": as many charges as possible are chosen, each with size 1
          (or more if the total number of charges is less than the total size).
        - "minimal": only the zero charge sector is chosen, with full size.
        - tuple: the sizes of the charge sectors, a matching number of charges
          are chosen automatically, in sequence 'closest to zero'.

    seed : None, int, or numpy.random.Generator, optional
        The seed for the random number generator.

    Returns
    -------
    BlockIndex
    """
    import symmray as sr

    if isinstance(d, sr.BlockIndex):
        # already a BlockIndex, nothing more to generate
        return d

    symmetry = sr.get_symmetry(symmetry)

    rng = get_rng(seed)

    if (dual is None) or (dual == "random"):
        dual = rng.choice([False, True])

    if isinstance(d, dict):
        # charges and sizes given explicitly, nothing more to generate
        return sr.BlockIndex(chargemap=d, dual=dual)

    if not isinstance(d, numbers.Integral):
        # charges and sizes given as a linearmap, nothing more to generate
        return sr.BlockIndex(linearmap=d, dual=dual)

    kwargs = {
        "dual": dual,
        "subsizes": subsizes,
        "seed": rng,
    }

    if isinstance(symmetry, sr.ZN):
        return rand_zn_index(d, order=symmetry.N, **kwargs)
    elif symmetry == "Z2Z2":
        return rand_z2z2_index(d, **kwargs)
    elif symmetry == "U1":
        return rand_u1_index(d, **kwargs)
    elif symmetry == "U1U1":
        return rand_u1u1_index(d, **kwargs)
    else:
        raise ValueError(f"Symmetry unknown or not supported: {symmetry}.")


def get_array_cls(symmetry, fermionic=False, flat=False) -> type:
    """Get the array class for the given symmetry, fermionic flag, and flat
    flag.

    Parameters
    ----------

    """
    import symmray as sr

    if symmetry in ("Z2", "U1", "Z2Z2", "U1U1"):
        # statically defined array classes
        return {
            # blocksparse abelian arrays
            ("Z2", 0, 0): sr.Z2Array,
            ("U1", 0, 0): sr.U1Array,
            ("Z2Z2", 0, 0): sr.Z2Z2Array,
            ("U1U1", 0, 0): sr.U1U1Array,
            # blocksparse fermionic arrays
            ("Z2", 1, 0): sr.Z2FermionicArray,
            ("U1", 1, 0): sr.U1FermionicArray,
            ("Z2Z2", 1, 0): sr.Z2Z2FermionicArray,
            ("U1U1", 1, 0): sr.U1U1FermionicArray,
            # flat abelian arrays
            ("Z2", 0, 1): sr.Z2ArrayFlat,
            # flat fermionic arrays
            ("Z2", 1, 1): sr.Z2FermionicArrayFlat,
        }[str(symmetry), fermionic, flat]
    else:
        # symmetry is defined dynamically, and should be supplied as kwarg
        return {
            (0, 0): sr.AbelianArray,
            (1, 0): sr.FermionicArray,
            (0, 1): sr.AbelianArrayFlat,
            (1, 1): sr.FermionicArrayFlat,
        }[fermionic, flat]


def choose_duals(duals, ndim):
    if duals == "equal":
        # split ~half and ~half
        return [i >= ndim // 2 for i in range(ndim)]
    elif (
        (duals == "random")
        or (duals is None)
        or (duals is False)
        or (duals is True)
    ):
        # repeat for all axes
        return [duals] * ndim
    else:
        # assume given explicit sequence
        if len(duals) != ndim:
            raise ValueError(
                f"Length of duals ({len(duals)}) does not match ndim ({ndim})."
            )
        return duals


def get_rand(
    symmetry,
    shape,
    duals="random",
    charge=None,
    seed=None,
    dist="normal",
    fermionic=False,
    label=None,
    flat=False,
    subsizes="random",
    **kwargs,
):
    """Get a random symmray array, with the given symmetry and shape. The
    duals, charge, and sub charge sizes can be specified or automatically
    or randomly generated.

    Parameters
    ----------
    symmetry : str
        The symmetry of the array.
    shape : tuple[int | dict | sequence | BlockIndex, ...]
        The desired overall effective shape of the array. Each element can be
        an int, in which case the charge sizes will be generated automatically
        according to `subsizes`, or an explicit dict of charge sizes, or an
        explicit linearmap, or a `BlockIndex`.
    duals : "random", "equals", or Sequence[bool], optional
        The dualness of each index. If "random", the dualnesses are chosen
        randomly. If "equal", they are chosen so the first half of the
        indices have `dual=False` and the second half have `dual=True`.
        If `shape` contains `BlockIndex` objects, the dualness of that index
        takes precedence over any specification here.
    charge : int or tuple[int], optional
        The total charge of the array. If not given the 'zero' charge is used,
        which depends on the symmetry.
    seed : None, int, or numpy.random.Generator, optional
        The seed for the random number generator.
    dist : str, optional
        The distribution of the random numbers. Can be "normal" or "uniform".
    fermionic : bool, optional
        Whether to generate a fermionic array.
    label : hashable, optional
        An optional label for the array, potentially needed for ordering dummy
        odd fermionic modes.
    flat : bool, optional
        Whether to generate a 'flat' backend array (True) or the default
        block-sparse backend array (False).
    subsizes : "random", "equal", "maximal", "minimal", or tuple[int], optional
        The sizes of the charge sectors. The choices are as follows:

        - "random": the charges and sizes are randomly determined.
        - "equal": a fixed number of charges 'close to' zero charge are chosen,
          all with equal size (up to remainders).
        - "maximal": as many charges as possible are chosen, each with size 1
          (or more if the total number of charges is less than the total size).
        - "minimal": only the zero charge sector is chosen, with full size.
        - tuple: the sizes of the charge sectors, a matching number of charges
          are chosen automatically, in sequence 'closest to zero'.

        If `shape` contains `dict` or `BlockIndex` objects, the subsizes of
        those indices take precedence over any specification here.
    kwargs
        Additional keyword arguments are passed to the random array
        generation function.

    Returns
    -------
    AbelianArray or FermionicArray
    """
    import symmray as sr

    symmetry = sr.get_symmetry(symmetry)
    rng = get_rng(seed)
    duals = choose_duals(duals, len(shape))
    cls = get_array_cls(symmetry, fermionic=fermionic, flat=False)

    indices = [
        rand_index(symmetry, d, dual=dual, subsizes=subsizes, seed=rng)
        for d, dual in zip(shape, duals)
    ]

    if not fermionic:
        kwargs.pop("oddpos", None)

    x = cls.random(
        indices=indices,
        charge=charge,
        seed=rng,
        dist=dist,
        symmetry=symmetry,
        label=label,
        **kwargs,
    )

    if flat:
        x = x.to_flat()

    return x


def get_rand_blockvector(
    size,
    block_size=0.25,
    seed=None,
    dist="normal",
):
    import symmray as sr

    rng = get_rng(seed)
    blocks = {}
    d = 0
    i = 0
    while d < size:
        if block_size < 1:
            # take as typical fraction
            block_size = rng.poisson(block_size * size)
        else:
            # take as absolute size
            block_size = int(block_size)

        block_size = min(max(block_size, 1), size - d)
        block = getattr(rng, dist)(size=block_size)
        blocks[i] = block
        d += block_size
        i += 1

    return sr.BlockVector(blocks)


def from_dense(
    array,
    symmetry,
    index_maps,
    duals=None,
    fermionic=False,
    charge=None,
):
    cls = get_array_cls(symmetry, fermionic=fermionic, flat=False)
    return cls.from_dense(
        array,
        index_maps,
        duals=duals,
        charge=charge,
        symmetry=symmetry,
    )



import pytest

import symmray as sr

from .test_flat_abelian_array import get_zn_blocksparse_flat_compat


@pytest.mark.parametrize("symmetry", ("Z2", "Z4"))
@pytest.mark.parametrize("seed", range(5))
@pytest.mark.parametrize("ndim", [1, 2, 3, 4, 5])
@pytest.mark.parametrize("all_axes", [False, True])
def test_sort_sectors(
    symmetry,
    seed,
    ndim,
    all_axes,
):
    rng = sr.utils.get_rng(seed)
    shape = [rng.choice([2, 4]) for _ in range(ndim)]
    if ndim > 2:
        num_sort = rng.integers(1, ndim)
    else:
        num_sort = 1
    axes_sort = tuple(rng.choice(ndim, size=num_sort, replace=False))
    x = get_zn_blocksparse_flat_compat(
        symmetry,
        shape=shape,
        charge=0,
        fermionic=True,
        seed=rng,
    )
    x.randomize_phases(seed + 1, inplace=True)
    fx: sr.FermionicArrayFlat = x.to_flat()
    fx.sort_stack(axes_sort, all_axes=all_axes, inplace=True)
    fx.check()
    y = fx.to_blocksparse()
    y.check()
    y.test_allclose(x)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("charge", [0, 1])
@pytest.mark.parametrize("seed", [42, 43, 44])
@pytest.mark.parametrize("sync", [False, True])
def test_to_and_from_blocksparse_with_phase_sync(
    symmetry,
    charge,
    seed,
    sync,
):
    if charge:
        pytest.xfail("oddpos not implemented yet.")

    x = get_zn_blocksparse_flat_compat(
        symmetry,
        (2, 4, 6),
        charge=charge,
        fermionic=True,
        seed=seed,
    )
    # add some non-trivial phases
    x.transpose((2, 0, 1), inplace=True)
    assert x.phases
    fx = x.to_flat()
    assert fx.fermionic
    if sync:
        fx.phase_sync(inplace=True)
    fx.check()
    y = fx.to_blocksparse()
    assert sync == (not y.phases)
    y.check()
    y.test_allclose(x)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("charge", [0, 1])
@pytest.mark.parametrize("seed", [42, 43, 44])
@pytest.mark.parametrize("sync", [False, True])
@pytest.mark.parametrize("axs", [(), (0,), (1,), (0, 2), (0, 2, 1)])
def test_phase_flip(
    symmetry,
    charge,
    seed,
    sync,
    axs,
):
    if charge:
        pytest.xfail("oddpos not implemented yet.")

    x = get_zn_blocksparse_flat_compat(
        symmetry,
        (2, 4, 6),
        charge=charge,
        fermionic=True,
        seed=seed,
    )
    x.randomize_phases(seed + 1, inplace=True)
    assert x.phases
    fx = x.to_flat()
    xflipped = x.phase_flip(*axs)
    fxflipped = fx.phase_flip(*axs)
    if sync:
        fxflipped.phase_sync(inplace=True)
    fxflipped.check()
    y = fxflipped.to_blocksparse()
    y.check()
    if sync:
        # phases should have been absorbed into blocks
        assert not y.phases
    y.test_allclose(xflipped)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("charge", [0, 1])
@pytest.mark.parametrize("seed", range(20))
@pytest.mark.parametrize("sync", [False, True])
def test_phase_transpose(
    symmetry,
    charge,
    seed,
    sync,
):
    if charge:
        pytest.xfail("oddpos not implemented yet.")

    rng = sr.utils.get_rng(seed)
    N = rng.integers(1, 7)
    perm = tuple(rng.permutation(N))
    shape = tuple(rng.choice([2, 4], size=N))

    x = get_zn_blocksparse_flat_compat(
        symmetry,
        shape,
        charge=charge,
        fermionic=True,
        seed=seed,
    )
    # add some non-trivial phases
    x.randomize_phases(seed + 1, inplace=True)
    fx = x.to_flat()

    x_phase_transposed = x.phase_transpose(perm)
    fx_phase_transposed = fx.phase_transpose(perm)
    if sync:
        fx_phase_transposed.phase_sync(inplace=True)
    fx_phase_transposed.check()
    y = fx_phase_transposed.to_blocksparse()
    y.check()
    if sync:
        # phases should have been absorbed into blocks
        assert not y.phases
    y.test_allclose(x_phase_transposed)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("charge", [0, 1])
@pytest.mark.parametrize("seed", [42, 43, 44])
def test_phase_global(symmetry, charge, seed):
    if charge:
        pytest.xfail("oddpos not implemented yet.")
    x = get_zn_blocksparse_flat_compat(
        symmetry,
        (2, 4, 6),
        charge=charge,
        fermionic=True,
        seed=seed,
    )
    # add some non-trivial phases
    x.randomize_phases(seed + 1, inplace=True)
    fx = x.to_flat()
    fxg = fx.phase_global()
    fxg.to_blocksparse().test_allclose(x.phase_global())


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("charge", [0, 1])
@pytest.mark.parametrize("ndim", [1, 2, 3, 4, 5])
@pytest.mark.parametrize("seed", range(10))
@pytest.mark.parametrize("sync", [False, True])
def test_transpose(
    symmetry,
    charge,
    ndim,
    seed,
    sync,
):
    if charge:
        pytest.xfail("oddpos not implemented yet.")

    rng = sr.utils.get_rng(seed)

    shape = [rng.choice([2, 4]) for _ in range(ndim)]

    if rng.random() < 0.2:
        perm = None
    else:
        perm = rng.permutation(ndim)

    x = get_zn_blocksparse_flat_compat(
        symmetry,
        shape,
        charge=charge,
        fermionic=True,
        seed=rng,
    )
    x.randomize_phases(rng, inplace=True)
    xt = x.transpose(perm)
    fx = x.to_flat()
    fxt = fx.transpose(perm)
    fxt.check()
    if sync:
        fxt.phase_sync(inplace=True)
    y = fxt.to_blocksparse()
    y.check()
    y.test_allclose(xt)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("shape", [(2, 6, 4), (2, 2, 2, 2)])
@pytest.mark.parametrize("charge", [0, 1])
@pytest.mark.parametrize("seed", [42, 43, 44])
@pytest.mark.parametrize("sync", [False, True])
@pytest.mark.parametrize("phase_permutation", [False, True])
@pytest.mark.parametrize("phase_dual", [False, True])
def test_conj(
    symmetry,
    shape,
    charge,
    seed,
    sync,
    phase_permutation,
    phase_dual,
):
    x = get_zn_blocksparse_flat_compat(
        symmetry,
        shape,
        charge=charge,
        fermionic=True,
        oddpos="x",
        seed=seed,
    )
    # add some non-trivial phases
    x.randomize_phases(seed + 1, inplace=True)
    assert x.phases
    xc = x.conj(phase_permutation=phase_permutation, phase_dual=phase_dual)

    fx = x.to_flat()
    fxc = fx.conj(phase_permutation=phase_permutation, phase_dual=phase_dual)
    fxc.check()
    if sync:
        fxc.phase_sync(inplace=True)
    y = fxc.to_blocksparse()
    y.check()
    y.test_allclose(xc)


@pytest.mark.parametrize("symmetry", ("Z2", "Z4"))
@pytest.mark.parametrize("charge", [0, 1])
@pytest.mark.parametrize("ndim", [1, 2, 3, 4, 5])
@pytest.mark.parametrize("dtype", ["complex128", "float64"])
@pytest.mark.parametrize("seed", range(5))
def test_dagger(symmetry, charge, ndim, dtype, seed):
    rng = sr.utils.get_rng(seed)
    xs = get_zn_blocksparse_flat_compat(
        symmetry,
        shape=[2] * ndim,
        charge=charge,
        fermionic=True,
        oddpos="x",
        seed=rng,
        dtype=dtype,
    )
    xs.randomize_phases(seed + 1, inplace=True)
    x = xs.to_flat()
    x.dagger().test_allclose(x.H)
    x.dagger().test_allclose(x.conj().transpose())
    x.dagger().to_blocksparse().test_allclose(xs.dagger())


@pytest.mark.parametrize(
    "symmetry, charge",
    [
        ("Z2", 0),
        ("Z2", 1),
        ("Z4", 0),
        ("Z4", 1),
        ("Z4", 2),
        ("Z4", 3),
    ],
)
@pytest.mark.parametrize("seed", range(3))
def test_fermi_norm_phase_dual(symmetry, charge, seed):
    sx = sr.utils.get_rand(
        symmetry,
        (4, 8, 4, 8),
        subsizes="equal",
        fermionic=True,
        charge=charge,
        oddpos="x",
        flat=False,
        seed=seed,
    )
    x = sx.to_flat()
    n2 = x.norm() ** 2

    xd = x.dagger(phase_dual=True)
    assert xd.to_blocksparse().test_allclose(sx.dagger(phase_dual=True))
    na = sr.tensordot(xd, x, axes=[(3, 2, 1, 0), (0, 1, 2, 3)])
    nb = sr.tensordot(x, xd, axes=[(3, 2, 1, 0), (0, 1, 2, 3)])
    assert na == pytest.approx(nb)
    assert na == pytest.approx(n2)
    assert nb == pytest.approx(n2)

    xc = x.conj(phase_dual=True)
    assert xc.to_blocksparse().test_allclose(sx.conj(phase_dual=True))
    nc = sr.tensordot(xc, x, axes=[(0, 1, 2, 3), (0, 1, 2, 3)])
    nd = sr.tensordot(x, xc, axes=[(0, 1, 2, 3), (0, 1, 2, 3)])
    assert nc == pytest.approx(nd)
    assert nc == pytest.approx(n2)
    assert nd == pytest.approx(n2)

    xct = x.conj(phase_dual=True).transpose()
    assert xct.to_blocksparse().test_allclose(
        sx.conj(phase_dual=True).transpose()
    )
    ne = sr.tensordot(xct, x, axes=[(3, 2, 1, 0), (0, 1, 2, 3)])
    nf = sr.tensordot(x, xct, axes=[(3, 2, 1, 0), (0, 1, 2, 3)])
    assert ne == pytest.approx(nf)
    assert ne == pytest.approx(n2)
    assert nf == pytest.approx(n2)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("charge", [0, 1])
@pytest.mark.parametrize(
    "shape,axes_groups",
    [
        ((2, 4, 6, 4), ((0,), (1,), (2,), (3,))),
        ((2, 4, 6, 4), ()),
        ((2, 4, 6, 4), ((0, 1),)),
        ((2, 4, 6, 4), ((1, 2),)),
        ((2, 4, 6, 4), ((1, 3),)),
        ((2, 4, 6, 4), ((2, 3),)),
        ((2, 4, 6, 4), ((0, 1), (2, 3))),
        ((2, 4, 6, 4), ((0, 2), (1, 3))),
        ((2, 4, 6, 4), ((0, 3), (1, 2))),
        ((2, 4, 6, 4), ((0, 3, 1, 2),)),
        ((2, 4, 6, 4, 6), ((0, 1), (2, 3, 4))),
        ((2, 4, 6, 4, 6), ((0, 2), (1, 3, 4))),
        ((2, 4, 6, 4, 6), ((0, 4), (1, 2, 3))),
        ((2, 4, 6, 4, 6), ((0, 1, 2), (3, 4))),
        ((2, 4, 6, 4, 6), ((0, 1, 3), (2, 4))),
        ((2, 4, 6, 4, 6), ((0, 1, 4), (2, 3))),
    ],
)
def test_fuse(
    symmetry,
    charge,
    shape,
    axes_groups,
):
    if charge:
        pytest.xfail("oddpos not implemented yet.")

    x = get_zn_blocksparse_flat_compat(
        symmetry,
        shape=shape,
        charge=charge,
        fermionic=True,
        seed=42,
    )
    x.randomize_phases(43, inplace=True)
    x_fused = x.fuse(*axes_groups, inplace=False)
    fx = x.to_flat()
    fx_fused = fx.fuse(*axes_groups, inplace=False)
    fx_fused.check()
    y = fx_fused.to_blocksparse()
    y.check()
    y.test_allclose(x_fused)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("charge", [0, 1])
@pytest.mark.parametrize("seed", range(10))
def test_fuse_unfuse(symmetry, charge, seed):
    if charge:
        pytest.xfail("oddpos not implemented yet.")

    rng = sr.utils.get_rng(seed)
    x = get_zn_blocksparse_flat_compat(
        symmetry,
        shape=(2, 4, 6, 4, 8),
        charge=charge,
        fermionic=True,
        seed=rng,
    )
    x.randomize_phases(seed + 1, inplace=True)
    fx = x.to_flat()

    nfuse = rng.integers(1, x.ndim)
    axes = tuple(rng.choice(x.ndim, size=nfuse, replace=False))
    position = min(axes)
    new_order = (
        *range(position),
        *axes,
        *(ax for ax in range(position, x.ndim) if ax not in axes),
    )
    perm_back = sorted(range(x.ndim), key=lambda i: new_order[i])

    x_fused = x.fuse(axes)
    fx_fused = fx.fuse(axes)
    fx_fused.check()
    fx_fused.to_blocksparse().test_allclose(x_fused)

    if len(axes) > 1:
        y = x_fused.unfuse(position)
        fy = fx_fused.unfuse(position)
        fy.check()
        fy.to_blocksparse().test_allclose(y)
    else:
        y = x_fused
        fy = fx_fused

    fyt = fy.transpose(perm_back)
    fyt.check()
    fyt.to_blocksparse().test_allclose(x)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize(
    "shape,axes_groups",
    [
        ([4], [(0,)]),
        ([2, 2], [(0, 1)]),
        ([2, 2], [(0,), (1,)]),
        ([2, 2], [(1,), (0,)]),
        ([4] * 6, [(1, 3), (4, 2)]),
        ([6, 2, 4, 8], [(3, 2, 1)]),
        ([6, 2, 4, 8], [(0, 1)]),
        ([6, 2, 4, 8], [(2, 3)]),
        ([6, 2, 4, 8], [(0, 1, 2, 3)]),
        ([6, 2, 4, 8], [(2, 3, 1, 0)]),
        ([6, 2, 4, 8], [(0, 1), (2, 3)]),
        ([2, 2, 2, 2, 2], [(0, 1), (2, 3, 4)]),
        ([2, 2, 2, 2, 2, 2], [(0, 1), (2, 3), (4, 5)]),
        ([2, 2, 2, 2, 2, 2], [(0, 1), (4, 5), (2, 3)]),
        ([4, 2, 6, 2], [(0, 3)]),
        ([2, 2, 2, 2, 2, 2, 2, 2], [(5,), (7, 2, 3), (1, 4)]),
    ],
)
@pytest.mark.parametrize("charge", [0, 1])
def test_fuse_roundtrip(symmetry, shape, axes_groups, charge):
    if charge:
        pytest.xfail("oddpos not implemented yet.")

    sx = get_zn_blocksparse_flat_compat(
        symmetry, shape, charge, fermionic=True, seed=42
    )
    sx.randomize_phases(43, inplace=True)
    sy = sx.fuse(*axes_groups)
    fx = sx.to_flat()
    fx.check()
    fx.to_blocksparse().test_allclose(sx)
    fy = fx.fuse(*axes_groups)
    fy.check()
    fy.to_blocksparse().test_allclose(sy)
    xu = fy.unfuse_all()
    xu.check()
    # fuse + unfuse is identity up to permutation of axes
    axes_grouped = [i for g in axes_groups for i in g]
    axes_rem = [i for i in range(fx.ndim) if i not in axes_grouped]
    ax_g0 = min(axes_grouped)
    new_axes = axes_rem[:ax_g0] + axes_grouped + axes_rem[ax_g0:]
    sxt = sx.transpose(new_axes)
    fxus = xu.to_blocksparse()
    fxus.test_allclose(sxt)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("seed", range(50))
def test_tensordot(symmetry, seed):
    N = int(symmetry[1:])

    a, b, axes = sr.utils_test.rand_valid_tensordot(
        symmetry=symmetry,
        fermionic=True,
        dimension_multiplier=N,
        subsizes="equal",
        seed=seed,
    )
    a.randomize_phases(seed + 1, inplace=True)
    b.randomize_phases(seed + 2, inplace=True)
    c = a.tensordot(b, axes=axes, preserve_array=True)

    fa = a.to_flat()
    fb = b.to_flat()
    fc = fa.tensordot(fb, axes=axes, preserve_array=True)
    fc.check()

    if c.is_zero() and fc.is_zero():
        # both are zero, other tests might break
        return

    fc.to_blocksparse().test_allclose(c)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("ndim_a", [1, 2])
@pytest.mark.parametrize("ndim_b", [1, 2])
@pytest.mark.parametrize("charge_a", [0, 1])
@pytest.mark.parametrize("charge_b", [0, 1])
@pytest.mark.parametrize("seed", range(10))
def test_matmul(symmetry, ndim_a, ndim_b, charge_a, charge_b, seed):
    rng = sr.utils.get_rng(seed)
    da = rng.choice([12, 24, 36])
    db = rng.choice([12, 24, 36])
    dc = rng.choice([12, 24, 36])
    a = sr.utils.rand_index(symmetry, da, subsizes="equal", seed=rng)
    b = sr.utils.rand_index(symmetry, db, subsizes="equal", seed=rng)
    c = sr.utils.rand_index(symmetry, dc, subsizes="equal", seed=rng)
    if ndim_a == 1:
        shape_a = [b]
    else:
        shape_a = [a, b]
    x = sr.utils.get_rand(
        symmetry,
        shape_a,
        seed=rng,
        charge=charge_a,
        fermionic=True,
        oddpos="x",
    )
    x.randomize_phases(rng, inplace=True)
    if ndim_b == 1:
        shape_b = [b.conj()]
    else:
        shape_b = [b.conj(), c]
    y = sr.utils.get_rand(
        symmetry,
        shape_b,
        seed=rng,
        charge=charge_b,
        fermionic=True,
        oddpos="y",
    )
    y.randomize_phases(rng, inplace=True)
    z = x @ y
    fx = x.to_flat()
    fy = y.to_flat()
    fz = fx @ fy

    if hasattr(fz, "check"):
        fz.check()
        fz.to_blocksparse().test_allclose(z)
    else:
        # scalar
        assert fz == pytest.approx(z)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("charge", [0, 1])
@pytest.mark.parametrize("axis", [0, 1, 2, 3])
def test_block_multiply_diagonal(symmetry, charge, axis):
    import autoray as ar
    import numpy as np

    if charge:
        pytest.xfail("oddpos not implemented yet.")

    rng = np.random.default_rng(42)
    sx = get_zn_blocksparse_flat_compat(
        symmetry,
        (4, 2, 6, 2),
        seed=rng,
        charge=charge,
        fermionic=True,
    )
    sx.randomize_phases(rng, inplace=True)
    x = sx.to_flat()

    v = sr.FlatVector.rand(symmetry, x.indices[axis].charge_size)
    y = ar.do("multiply_diagonal", x, v, axis=axis)

    # check dense reference
    xd = x.to_dense()
    vd = v.to_dense()
    yd = y.to_dense()
    lhs = "abcd"
    rhs = lhs[axis]
    np.testing.assert_allclose(yd, np.einsum(f"{lhs},{rhs}->{lhs}", xd, vd))


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("seed", range(5))
@pytest.mark.parametrize("fn", ["ldmul", "lddiv"])
def test_ldmul_and_lddiv(symmetry, seed, fn):
    import autoray as ar

    rng = sr.utils.get_rng(seed)
    da = rng.choice([12, 24, 36])
    db = rng.choice([12, 24, 36])
    shape_x = (da, db)
    sx = get_zn_blocksparse_flat_compat(
        symmetry,
        shape_x,
        charge=0,
        fermionic=True,
        seed=rng,
    )
    sx.randomize_phases(rng, inplace=True)
    vblocks = {
        c: rng.standard_normal(dc) + 1j * rng.standard_normal(dc)
        for c, dc in sx.indices[0].chargemap.items()
    }
    sv = sr.BlockVector(vblocks)
    sy = ar.do(fn, sv, sx)
    fx = sx.to_flat()
    fv = sv.to_flat()
    fy = ar.do(fn, fv, fx)
    fy.check()
    fy.to_blocksparse().test_allclose(sy)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("seed", range(5))
@pytest.mark.parametrize("fn", ["rdmul", "rddiv"])
def test_rdmul_and_rddiv(symmetry, seed, fn):
    import autoray as ar

    rng = sr.utils.get_rng(seed)
    da = rng.choice([12, 24, 36])
    db = rng.choice([12, 24, 36])
    shape_x = (da, db)
    sx = get_zn_blocksparse_flat_compat(
        symmetry,
        shape_x,
        charge=0,
        fermionic=True,
        seed=rng,
    )
    sx.randomize_phases(rng, inplace=True)
    vblocks = {
        c: rng.standard_normal(dc) + 1j * rng.standard_normal(dc)
        for c, dc in sx.indices[1].chargemap.items()
    }
    sv = sr.BlockVector(vblocks)
    sy = ar.do(fn, sx, sv)
    fx = sx.to_flat()
    fv = sv.to_flat()
    fy = ar.do(fn, fx, fv)
    fy.check()
    fy.to_blocksparse().test_allclose(sy)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("seed", range(2))
@pytest.mark.parametrize("ndim", [1, 2, 3, 4, 5])
def test_norm(symmetry, ndim, seed):
    rng = sr.utils.get_rng(seed)
    shape = [rng.choice([2, 4]) for _ in range(ndim)]
    x = get_zn_blocksparse_flat_compat(
        symmetry,
        shape=shape,
        charge=0,
        fermionic=True,
        seed=rng,
    )
    x.randomize_phases(seed + 1, inplace=True)
    n1 = x.norm()
    fx = x.to_flat()
    n2 = fx.norm()
    assert n1 == pytest.approx(n2)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("seed", range(2))
@pytest.mark.parametrize("dtype", ["complex128", "float64"])
@pytest.mark.parametrize("duals", [(False, True), (True, False)])
def test_eigh_fermionic_flat(symmetry, seed, dtype, duals):
    sx = get_zn_blocksparse_flat_compat(
        symmetry,
        (4, 4),
        fermionic=True,
        seed=seed,
        duals=duals,
        dtype=dtype,
    )
    sx.randomize_phases(seed + 1, inplace=True)
    # needs to be hermitian for eigh
    for sector, block in sx.get_sector_block_pairs():
        sx.set_block(sector, block + block.conj().T)

    sw, sU = sr.linalg.eigh(sx)
    sy = sr.multiply_diagonal(sU, sw, 1) @ sU.H
    sy.test_allclose(sx)

    fx = sx.to_flat()
    w, U = sr.linalg.eigh(fx)
    w.check()
    U.check()

    assert U.to_blocksparse().test_allclose(sU)

    y = sr.multiply_diagonal(U, w, 1) @ U.H
    y.test_allclose(fx)


@pytest.mark.parametrize("symmetry", ["Z2", "Z4"])
@pytest.mark.parametrize("seed", range(2))
@pytest.mark.parametrize("dtype", ["complex128", "float64"])
@pytest.mark.parametrize("duals", [(False, True), (True, False)])
def test_eigh_truncated_fermionic_flat(symmetry, seed, dtype, duals):
    sx = get_zn_blocksparse_flat_compat(
        symmetry,
        (4, 4),
        fermionic=True,
        seed=seed,
        duals=duals,
        dtype=dtype,
    )
    sx.randomize_phases(seed + 1, inplace=True)
    # needs to be hermitian for eigh
    for sector, block in sx.get_sector_block_pairs():
        sx.set_block(sector, block + block.conj().T)

    fx = sx.to_flat()
    u, w, uh = sr.linalg.eigh_truncated(fx, max_bond=-1, absorb=None)
    y = sr.multiply_diagonal(u, w, 1) @ uh
    y.test_allclose(fx)

    z = u @ u.dagger_project_left() @ fx @ uh.dagger_project_right() @ uh
    z.test_allclose(fx)



"""Interface to linear algebra submodule functions."""

import autoray as ar


def eigh(x, *args, **kwargs):
    """Hermitian eigen-decomposition of an assumed hermitian symmray array.

    Returns
    -------
    w : VectorCommon
        The eigenvalues as a vector.
    u : AbelianCommon
        The array of eigenvectors.
    """
    return x.eigh(*args, **kwargs)


def eigh_truncated(x, *args, **kwargs):
    """Truncated hermitian eigen-decomposition of an assumed hermitian symmray
    array.

    Parameters
    ----------
    cutoff : float, optional
        Absolute eigenvalue cutoff threshold.
    cutoff_mode : int or str, optional
        How to perform the truncation:

        - 1 or 'abs': trim values below ``cutoff``
        - 2 or 'rel': trim values below ``s[0] * cutoff``
        - 3 or 'sum2': trim s.t. ``sum(s_trim**2) < cutoff``.
        - 4 or 'rsum2': trim s.t. ``sum(s_trim**2) < sum(s**2) * cutoff``.
        - 5 or 'sum1': trim s.t. ``sum(s_trim**1) < cutoff``.
        - 6 or 'rsum1': trim s.t. ``sum(s_trim**1) < sum(s**1) * cutoff``.

    max_bond : int
        An explicit maximum bond dimension, use -1 for none.
    absorb : {-1, 0, 1, None}
        How to absorb the eigenvalues.

        - -1 or 'left': absorb into the left factor (U).
        - 0 or 'both': absorb the square root into both factors.
        - 1 or 'right': absorb into the right factor (VH).
        - None: do not absorb, return eigenvalues as a BlockVector.

    renorm : {0, 1}
        Whether to renormalize the eigenvalues (depends on `cutoff_mode`).

    Returns
    -------
    u : AbelianCommon
        The abelian array of left eigenvectors.
    w : VectorCommon or None
        The vector of eigenvalues, or None if absorbed.
    uh : AbelianCommon
        The abelian array of right eigenvectors.
    """
    return x.eigh_truncated(*args, **kwargs)


def norm(x, *args, **kwargs):
    return x.norm(*args, **kwargs)


def qr(x, *args, **kwargs):
    return x.qr(*args, **kwargs)


def qr_stabilized(x, *args, **kwargs):
    q, r = x.qr(*args, stabilized=True, **kwargs)
    return q, None, r


def solve(x, *args, **kwargs):
    return x.solve(*args, **kwargs)


def svd(x, *args, **kwargs):
    return x.svd(*args, **kwargs)


def svd_truncated(x, *args, **kwargs):
    """Truncated singular value decomposition of a symmray array.

    Parameters
    ----------
    cutoff : float, optional
        Singular value cutoff threshold.
    cutoff_mode : int or str, optional
        How to perform the truncation:

        - 1 or 'abs': trim values below ``cutoff``
        - 2 or 'rel': trim values below ``s[0] * cutoff``
        - 3 or 'sum2': trim s.t. ``sum(s_trim**2) < cutoff``.
        - 4 or 'rsum2': trim s.t. ``sum(s_trim**2) < sum(s**2) * cutoff``.
        - 5 or 'sum1': trim s.t. ``sum(s_trim**1) < cutoff``.
        - 6 or 'rsum1': trim s.t. ``sum(s_trim**1) < sum(s**1) * cutoff``.

    max_bond : int
        An explicit maximum bond dimension, use -1 for none.
    absorb : {-1, 0, 1, None}
        How to absorb the singular values.

        - -1 or 'left': absorb into the left factor (U).
        - 0 or 'both': absorb the square root into both factors.
        - 1 or 'right': absorb into the right factor (VH).
        - None: do not absorb, return singular values as a BlockVector.

    renorm : {0, 1}
        Whether to renormalize the singular values (depends on `cutoff_mode`).

    Returns
    -------
    u : AbelianCommon
        The abelian array of left singular vectors.
    s : VectorCommon or None
        The vector of singular values, or None if absorbed.
    vh : AbelianCommon
        The abelian array of right singular vectors.
    """
    return x.svd_truncated(*args, **kwargs)


# used by quimb
ar.register_function("symmray", "eigh_truncated", eigh_truncated)
ar.register_function("symmray", "qr_stabilized", qr_stabilized)
ar.register_function("symmray", "svd_truncated", svd_truncated)



import random

import pytest

import symmray as sr


def get_zn_blocksparse_flat_compat(
    symmetry,
    shape,
    charge=0,
    seed=42,
    shape_relative_to_z2=True,
    **kwargs,
):
    rng = sr.utils.get_rng(seed)

    N = int(symmetry[1:])

    if shape_relative_to_z2:
        shape = [N * d // 2 for d in shape]

    if charge:
        charge = rng.integers(low=1, high=N)

    return sr.utils.get_rand(
        symmetry,
        shape=shape,
        subsizes="equal",
        charge=charge,
        seed=seed,
        **kwargs,
    )


@pytest.mark.parametrize("symmetry", ["Z2", "Z3", "Z4"])
@pytest.mark.parametrize(
    "shape,perm",
    [
        ([2], [0]),
        ([4, 2], [0, 1]),
        ([2, 4], [1, 0]),
        ([2, 4, 2, 4], (2, 1, 3, 0)),
        ([2, 2, 2, 2, 2, 2, 2], (2, 0, 1, 3, 5, 4, 6)),
    ],
)
@pytest.mark.parametrize("charge", [0, 1])
def test_transpose_roundtrip(symmetry, shape, perm, charge):
    sx = get_zn_blocksparse_flat_compat(symmetry, shape, charge, seed=42)
    sy = sx.transpose(perm)
    perm_inv = sorted(range(len(perm)), key=lambda i: perm[i])

    fx = sx.to_flat()
    fx.check()
    assert fx.to_blocksparse().allclose(sx)
    fy = fx.transpose(perm)
    fy.check()
    assert fy.to_blocksparse().allclose(sy)
    fz = fy.transpose(perm_inv)
    fz.check()
    assert fz.to_blocksparse().allclose(sx)


@pytest.mark.parametrize("symmetry", ["Z2", "Z3", "Z4"])
@pytest.mark.parametrize(
    "shape,axes_groups",
    [
        ([4], [(0,)]),
        ([2, 2], [(0, 1)]),
        ([2, 2], [(0,), (1,)]),
        ([2, 2], [(1,), (0,)]),
        ([4] * 6, [(1, 3), (4, 2)]),
        ([6, 2, 4, 8], [(3, 2, 1)]),
        ([6, 2, 4, 8], [(0, 1)]),
        ([6, 2, 4, 8], [(2, 3)]),
        ([6, 2, 4, 8], [(0, 1, 2, 3)]),
        ([6, 2, 4, 8], [(2, 3, 1, 0)]),
        ([6, 2, 4, 8], [(0, 1), (2, 3)]),
        ([2, 2, 2, 2, 2], [(0, 1), (2, 3, 4)]),
        ([2, 2, 2, 2, 2, 2], [(0, 1), (2, 3), (4, 5)]),
        ([2, 2, 2, 2, 2, 2], [(0, 1), (4, 5), (2, 3)]),
        ([4, 2, 6, 2], [(0, 3)]),
        ([2, 2, 2, 2, 2, 2, 2, 2], [(5,), (7, 2, 3), (1, 4)]),
    ],
)
@pytest.mark.parametrize("charge", [0, 1])
def test_fuse_roundtrip(symmetry, shape, axes_groups, charge):
    sx = get_zn_blocksparse_flat_compat(symmetry, shape, charge, seed=42)
    sy = sx.fuse(*axes_groups)
    fx = sx.to_flat()
    fx.check()
    assert fx.to_blocksparse().allclose(sx)
    fy = fx.fuse(*axes_groups)
    fy.check()
    assert fy.to_blocksparse().allclose(sy)
    xu = fy.unfuse_all()
    xu.check()
    # fuse + unfuse is identity up to permutation of axes
    axes_grouped = [i for g in axes_groups for i in g]
    axes_rem = [i for i in range(fx.ndim) if i not in axes_grouped]
    ax_g0 = min(axes_grouped)
    new_axes = axes_rem[:ax_g0] + axes_grouped + axes_rem[ax_g0:]
    sxt = sx.transpose(new_axes)
    fxus = xu.to_blocksparse()
    assert fxus.allclose(sxt)


@pytest.mark.parametrize("charge", [0, 1])
@pytest.mark.parametrize(
    "symmetry,shape,newshape",
    [
        ["Z2", (2, 2, 2, 2), (4, 2, 2)],
        ["Z2", (2, 2, 2, 2), (2, 4, 2)],
        ["Z2", (2, 2, 2, 2), (2, 2, 4)],
        ["Z2", (2, 2, 2, 2), (4, 4)],
        ["Z3", (3, 3, 3, 3), (9, 3, 3)],
        ["Z3", (3, 3, 3, 3), (3, 9, 3)],
        ["Z3", (3, 3, 3, 3), (3, 3, 9)],
        ["Z3", (3, 3, 3, 3), (9, 9)],
        ["Z4", (4, 4, 4, 4), (16, 4, 4)],
        ["Z4", (4, 4, 4, 4), (4, 16, 4)],
        ["Z4", (4, 4, 4, 4), (4, 4, 16)],
        ["Z4", (4, 4, 4, 4), (16, 16)],
        ["Z4", (4, 4, 4, 4, 4), (64, 4, 4)],
        ["Z4", (4, 4, 4, 4, 4), (4, 64, 4)],
        ["Z4", (4, 4, 4, 4, 4), (4, 4, 64)],
    ],
)
def test_reshape_roundtrip(symmetry, charge, shape, newshape):
    sx = get_zn_blocksparse_flat_compat(
        symmetry,
        shape,
        charge,
        seed=42,
        shape_relative_to_z2=False,
    )
    fx = sx.to_flat()
    sy = sx.reshape(newshape)
    fy = fx.reshape(newshape)
    fy.check()
    assert fy.to_blocksparse().allclose(sy)
    fz = fy.reshape(shape)
    fz.check()
    assert fz.to_blocksparse().allclose(sx)


@pytest.mark.parametrize("symmetry", ["Z2", "Z3", "Z4", "Z5"])
@pytest.mark.parametrize("seed", range(5))
@pytest.mark.parametrize("charge_x", [0, 1])
@pytest.mark.parametrize("charge_y", [0, 1])
def test_matmul(symmetry, seed, charge_x, charge_y):
    N = int(symmetry[1:])
    rng = sr.utils.get_rng(seed)

    da = rng.integers(1, 5)
    db = rng.integers(1, 5)
    dc = rng.integers(1, 5)

    a_ind = sr.utils.rand_index(symmetry, N * da, subsizes="equal", seed=rng)
    b_ind = sr.utils.rand_index(symmetry, N * db, subsizes="equal", seed=rng)
    c_ind = sr.utils.rand_index(symmetry, N * dc, subsizes="equal", seed=rng)

    if charge_x:
        charge_x = rng.integers(1, N)
    if charge_y:
        charge_y = rng.integers(1, N)

    sx = sr.utils.get_rand(
        symmetry, (a_ind, b_ind.conj()), charge=charge_x, seed=rng
    )
    sy = sr.utils.get_rand(symmetry, (b_ind, c_ind), charge=charge_y, seed=rng)
    sz = sx @ sy
    fx = sx.to_flat()
    fx.check()
    fy = sy.to_flat()
    fy.check()
    fz = fx @ fy
    fz.check()
    fz.to_blocksparse().allclose(sz)


@pytest.mark.parametrize("symmetry", ["Z2", "Z3", "Z4", "Z5"])
@pytest.mark.parametrize("seed", range(5))
@pytest.mark.parametrize("charge_x", [0, 1])
@pytest.mark.parametrize("charge_y", [0, 1])
def test_matvec(symmetry, seed, charge_x, charge_y):
    N = int(symmetry[1:])
    rng = sr.utils.get_rng(seed)

    da = rng.integers(1, 5)
    db = rng.integers(1, 5)

    a_ind = sr.utils.rand_index(symmetry, N * da, subsizes="equal", seed=rng)
    b_ind = sr.utils.rand_index(symmetry, N * db, subsizes="equal", seed=rng)

    if charge_x:
        charge_x = rng.integers(1, N)
    if charge_y:
        charge_y = rng.integers(1, N)

    sx = sr.utils.get_rand(
        symmetry, (a_ind, b_ind.conj()), charge=charge_x, seed=rng
    )
    sy = sr.utils.get_rand(symmetry, (b_ind,), charge=charge_y, seed=rng)
    sz = sx @ sy
    fx = sx.to_flat()
    fx.check()
    fy = sy.to_flat()
    fy.check()
    fz = fx @ fy
    fz.check()
    fz.to_blocksparse().allclose(sz)


@pytest.mark.parametrize("symmetry", ["Z2", "Z3", "Z4", "Z5"])
@pytest.mark.parametrize("seed", range(5))
@pytest.mark.parametrize("charge_x", [0, 1])
@pytest.mark.parametrize("charge_y", [0, 1])
def test_vecmat(symmetry, seed, charge_x, charge_y):
    N = int(symmetry[1:])
    rng = sr.utils.get_rng(seed)

    da = rng.integers(1, 5)
    db = rng.integers(1, 5)

    a_ind = sr.utils.rand_index(symmetry, N * da, subsizes="equal", seed=rng)
    b_ind = sr.utils.rand_index(symmetry, N * db, subsizes="equal", seed=rng)

    if charge_x:
        charge_x = rng.integers(1, N)
    if charge_y:
        charge_y = rng.integers(1, N)

    sx = sr.utils.get_rand(symmetry, (a_ind,), charge=charge_x, seed=rng)
    sy = sr.utils.get_rand(
        symmetry, (a_ind.conj(), b_ind), charge=charge_y, seed=rng
    )
    sz = sx @ sy
    fx = sx.to_flat()
    fx.check()
    fy = sy.to_flat()
    fy.check()
    fz = fx @ fy
    fz.check()
    fz.to_blocksparse().allclose(sz)


@pytest.mark.parametrize("ndim", [1, 2, 3, 4])
@pytest.mark.parametrize("order", [2, 3, 4])
@pytest.mark.parametrize("seed", range(5))
def test_build_cyclic_keys_conserve(ndim, order, seed):
    import numpy as np

    from symmray.flat.flat_array_common import (
        build_cyclic_keys_conserve,
        lexsort_sectors,
        zn_combine,
    )

    rng = random.Random(seed)
    charge = rng.randint(0, order - 1)
    duals = [rng.choice([True, False]) for _ in range(ndim)]

    sectors = build_cyclic_keys_conserve(
        ndim,
        order=order,
        charge=charge,
        duals=duals,
        flat=True,
    )
    scharges = zn_combine(order, sectors, duals=duals)

    assert set(map(int, scharges)) == {charge}
    assert np.all(lexsort_sectors(sectors) == np.arange(order ** (ndim - 1)))


@pytest.mark.parametrize("ndim", [1, 2, 3, 4])
@pytest.mark.parametrize("order", [2, 3, 4])
@pytest.mark.parametrize("seed", range(5))
def test_build_cyclic_keys_by_charge(ndim, order, seed):
    import numpy as np

    from symmray.flat.flat_array_common import (
        build_cyclic_keys_by_charge,
        lexsort_sectors,
        zn_combine,
    )

    rng = random.Random(seed)
    duals = [rng.choice([True, False]) for _ in range(ndim)]

    sectors = build_cyclic_keys_by_charge(
        ndim,
        order=order,
        duals=duals,
    )
    scharges = zn_combine(order, sectors, duals=duals)

    for i in range(order):
        # all have matching charge
        assert np.all(scharges[i] == i)
        # and are sorted within that charge
        assert np.all(
            lexsort_sectors(sectors[i]) == np.arange(order ** (ndim - 1))
        )


@pytest.mark.parametrize("symmetry", ["Z2", "Z3", "Z4"])
@pytest.mark.parametrize("seed", range(50))
def test_tensordot(symmetry, seed):
    from symmray.utils_test import rand_valid_tensordot

    N = int(symmetry[1:])

    a, b, axes = rand_valid_tensordot(
        symmetry,
        seed=seed,
        dimension_multiplier=N,
        subsizes="equal",
    )
    c = sr.tensordot(a, b, axes, preserve_array=True)

    fa = a.to_flat()
    fb = b.to_flat()

    fc = sr.tensordot(
        fa,
        fb,
        axes,
        preserve_array=True,
    )
    fc.check()

    if c.is_zero() and fc.is_zero():
        # both are zero, other tests might break
        return

    assert (
        len(
            set(map(float, fc.blocks.round(9).flat))
            - set(map(float, c.to_flat().blocks.round(9).flat))
        )
        == 0
    )

    fc.to_blocksparse().test_allclose(c)


@pytest.mark.parametrize("symm", ["Z2", "Z4"])
def test_tensordot_fused_with_already_fused_arrays(symm):
    a, b, c, d = (
        sr.utils.rand_index(symm, d, subsizes="equal") for d in [4, 4, 4, 8]
    )
    x = sr.utils.get_rand(symm, (a, b, c), flat=True)
    y = sr.utils.get_rand(symm, (c.conj(), d), flat=True)
    z = sr.tensordot(x, y, axes=[(2,), (0,)]).fuse((0, 1))
    xf = x.fuse((0, 1))
    zf = sr.tensordot(xf, y, axes=[(1,), (0,)])
    zf.test_allclose(z)


@pytest.mark.parametrize("symmetry", ["Z2", "Z3", "Z4"])
@pytest.mark.parametrize("charge", [0, 1])
@pytest.mark.parametrize("axis", [0, 1, 2, 3])
def test_block_multiply_diagonal(symmetry, charge, axis):
    import autoray as ar
    import numpy as np

    rng = np.random.default_rng(42)
    x = get_zn_blocksparse_flat_compat(
        symmetry,
        (4, 2, 6, 2),
        seed=rng,
        charge=charge,
    ).to_flat()

    v = sr.FlatVector.rand(symmetry, x.indices[axis].charge_size)
    y = ar.do("multiply_diagonal", x, v, axis=axis)

    # check dense reference
    xd = x.to_dense()
    vd = v.to_dense()
    yd = y.to_dense()
    lhs = "abcd"
    rhs = lhs[axis]
    np.testing.assert_allclose(yd, np.einsum(f"{lhs},{rhs}->{lhs}", xd, vd))


@pytest.mark.parametrize("symmetry", ["Z2", "Z3", "Z4"])
@pytest.mark.parametrize("ndim", [1, 2, 3, 4])
@pytest.mark.parametrize("seed", range(10))
def test_abelian_array_slice(symmetry, ndim, seed):
    import numpy as np

    rng = sr.utils.get_rng(seed)
    sx = get_zn_blocksparse_flat_compat(
        symmetry=symmetry,
        shape=tuple(2 * rng.integers(1, 5, size=ndim)),
        seed=rng,
    )
    dx = sx.to_dense()
    x = sx.to_flat()

    for ax in range(ndim):
        for d in range(x.shape[ax]):
            # get [:, :, ..., 2, ..., :] etc.
            selector = (
                (slice(None),) * ax + (d,) + (slice(None),) * (ndim - ax - 1)
            )
            x_slc = x[selector]
            x_slc.to_blocksparse().test_allclose(sx[selector])
            dx_slc = dx[selector]

            if ndim == 2:
                # dropped zero blocks will be fully missing
                dx_slc = dx_slc[dx_slc != 0]

            np.testing.assert_allclose(x_slc.to_dense(), dx_slc)
